/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth;

import javax.servlet.http.HttpServletRequest;

import at.gv.egiz.moodle.eidconnector.EIDConnectorException;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributes;

/**
 * Member state specific evaluation of credentials. First the user has to be redirected to the
 * memberstate specific authentication service. After successfully completing authentication the
 * member state specific authentication service is in posession of the user's credentials. Now the
 * authentication service has to invoke the moodle eId connector by redirecting the user to the
 * endpoint {@code http[s]://moodle-server.xyz/moodle-eid-connector/connector.do?action=authenticate}.
 * For security reasons the authentication framework should provide some kind of token that can be
 * used by the respective authentication provider class to retrieve the credentials from the
 * authentication service.
 * Upon invokation, the Moodle eId connector automatically iterates over all registered classes
 * (implementing this authentication provider interface). The {@code isResponsible(...)} method
 * checks if the request (providing the token...) is suitable. If suitable the second method,
 * {@code authenticate(...)} is automatically invoked. This method opens a communication channel
 * to the respective authentication service, using the token (or another mechanism) to retrieve
 * the credentials. The credentials have to be put in a {@link SubjectAttributes} object which is
 * returned by the method. 
 * {@link HttpServletRequest} originating from the respective authentication service.
 * @author AT-TUG (Thomas Knall)
 */
public interface AuthenticationProvider {
   
   /**
    * Returns {@code true} if the underlying authentication implementation claims to be responsible
    * for the given {@link HttpServletRequest}.
    * @param request The {@link HttpServletRequest}.
    * @return {@code true} if responsible, {@code false} if not.
    * @throws EIDConnectorException Thrown in case of an error.
    */
   boolean isResponsible(HttpServletRequest request) throws EIDConnectorException;

   /**
    * Uses the given request and retrieves/derives credentials from the respective authentication
    * service.
    * @param request The {@link HttpServletRequest}.
    * @return Credentials as {@link SubjectAttributes}.
    * @throws EIDConnectorException Thrown in case of an error.
    */
   SubjectAttributes authenticate(HttpServletRequest request) throws EIDConnectorException;
   
}
