/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.data;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.time.DateFormatUtils;
import org.apache.commons.lang.time.DateUtils;
import org.apache.log4j.Logger;

import at.gv.egiz.moodle.eidconnector.EIDConnectorException;
import at.gv.egiz.moodle.eidconnector.ErrorCode;
import at.iaik.commons.util.ToStringUtil;
import at.iaik.commons.xmlschema.saml.AgeOverData;

/**
 * @author AT-TUG (Thomas Knall)
 */
public class SubjectAttributes {
   
   public static final String EIDENTIFIER                  = "http://www.stork.gov.eu/1.0/eIdentifier";
   public static final String GIVEN_NAME                   = "http://www.stork.gov.eu/1.0/givenName";
   public static final String SURNAME                      = "http://www.stork.gov.eu/1.0/surname";
   public static final String INHERITATED_FAMILY_NAME      = "http://www.stork.gov.eu/1.0/inheritatedFamilyName";
   public static final String ADOPTED_FAMILY_NAME          = "http://www.stork.gov.eu/1.0/adoptedFamilyName";
   public static final String GENDER                       = "http://www.stork.gov.eu/1.0/gender";
   public static final String DATE_OF_BIRTH                = "http://www.stork.gov.eu/1.0/dateOfBirth";
   public static final String COUNTRY_OF_BIRTH             = "http://www.stork.gov.eu/1.0/countryCodeOfBirth";
   public static final String NATIONALITY                  = "http://www.stork.gov.eu/1.0/nationalityCode";
   public static final String MARITAL_STATUS               = "http://www.stork.gov.eu/1.0/maritalStatus";
   public static final String TEXT_RESIDIENCE_ADDRESS      = "http://www.stork.gov.eu/1.0/textResidenceAddress";
   public static final String CANONICAL_RESIDIENCE_ADDRESS = "http://www.stork.gov.eu/1.0/canonicalResidenceAddress";
   public static final String EMAIL_ADDRESS                = "http://www.stork.gov.eu/1.0/eMail";
   public static final String TITLE                        = "http://www.stork.gov.eu/1.0/title";
   public static final String RESIDIENCE_PERMIT            = "http://www.stork.gov.eu/1.0/residencePermit";
   public static final String PSEUDONYM                    = "http://www.stork.gov.eu/1.0/pseudonym";
   public static final String AGE                          = "http://www.stork.gov.eu/1.0/age";
   public static final String IS_AGE_OVER                  = "http://www.stork.gov.eu/1.0/isAgeOver";
   public static final String SIGNED_DOC                   = "http://www.stork.gov.eu/1.0/signedDoc";
   public static final String CITIZEN_X509CERTIFICATE      = "http://www.stork.gov.eu/1.0/citizen509Certificate";
   public static final String CITIZEN_QAA_LEVEL            = "http://www.stork.gov.eu/1.0/citizenQAAlevel";

   /**
    * @deprecated Use {@link #TITLE} instead.
    */
   public static final String ACADEMIC_TITLE               = "http://www.stork.gov.eu/1.0/academicTitle";

   /**
    * @deprecated Use {@link #CITIZEN_QAA_LEVEL} instead.
    */
   public static final String QAA_LEVEL                    = "http://www.stork.gov.eu/1.0/qaaLevel";
   
   /**
    * The eQualifier can be used in order to give further information on the eIdentifier.
    */
   public static final String EQUALIFIER              = "http://www.egiz.gv.at/1.0/eQualifier";
   
   public static final String[] ALL_STORK_ATTRIBUTES = new String[] {
      EIDENTIFIER,
      GIVEN_NAME,
      SURNAME,
      INHERITATED_FAMILY_NAME,
      ADOPTED_FAMILY_NAME,
      GENDER,
      DATE_OF_BIRTH,
      COUNTRY_OF_BIRTH,
      NATIONALITY,
      MARITAL_STATUS,
      TEXT_RESIDIENCE_ADDRESS,
      CANONICAL_RESIDIENCE_ADDRESS,
      EMAIL_ADDRESS,
      TITLE,
      RESIDIENCE_PERMIT,
      PSEUDONYM,
      AGE,
      IS_AGE_OVER,
      SIGNED_DOC,
      CITIZEN_X509CERTIFICATE,
      CITIZEN_QAA_LEVEL
   };
   
   private Hashtable<String, String> attributes;
   private Date timeStamp;
   private Date validAt;
   
   private List<AgeOverData> ageOverList;
   
   private Logger log = Logger.getLogger(SubjectAttributes.class);
   private boolean alreadyLogged;
   
   public SubjectAttributes() {
      this.attributes = new Hashtable<String, String>();
      this.timeStamp = null;
      this.validAt = null;
      this.ageOverList = new ArrayList<AgeOverData>();
      this.alreadyLogged = false;
   }
   
   public Set<String> getKeySet() {
      return Collections.unmodifiableSet(this.attributes.keySet());
   }

   protected Date getTimeStamp() {
      return this.timeStamp;
   }
   
   protected void setTimeStamp(Date timeStamp) {
      this.timeStamp = timeStamp;
   }

   public String get(String key) {
      return this.attributes.get(key);
   }
   
   public String set(String key, String value) {
      if (value == null) {
         return this.attributes.remove(key);
      }
      String oldValue = this.attributes.put(key, value);
      if (oldValue != null && SubjectAttributes.IS_AGE_OVER.equals(key)) {
         log.warn("Setting age over twice overwrites previously set values of age over. Use addIsAgeOver(AgeOverData) instead.");
      }
      return oldValue;
   }

   /**
    * Returns {@code true} if statistical logging has already been done so that no further logging is required.
    * @return {@code true} if statistical logging has already been done, {@code false} if not.
    */
   public boolean isAlreadyLogged() {
      return this.alreadyLogged;
   }

   /**
    * Should be set to {@code true} if statistical logging has already been performed by the
    * respective authentication implementation, {@code false} (which is the default value) if not.
    * @param alreadyLogged {@code true} if logging has already been performed.
    */
   public void setAlreadyLogged(boolean alreadyLogged) {
      this.alreadyLogged = alreadyLogged;
   }

   /**
    * The date the attributes encapsulated in this class are assured to be valid. May be needed
    * since there may be a delay in time between issuing of this attributes and retrieval by
    * underlying client application (e.g. Moodle). If set, this date will be used for age
    * calculation if needed, not not set ({@code null}) the current date will be used.
    * @return The point in time the attributes are regarded valid, or {@code null} if the current
    * date is to be used.
    */
   public Date getValidAt() {
      return this.validAt;
   }

   /**
    * Sets the date the attributes are guaranteed to be valid. Theoretically this date equals to the
    * current date, but in praxis some time may elapse until the attributes are retrieved by the
    * client application (some seconds up to the time frame the attributes are automaticalle removed
    * from the store). If not date is set the current date is taken for age calculation (if needed).
    * @param validAt The date the attributes are guaranteed to be valid.
    */
   public void setValidAt(Date validAt) {
      this.validAt = validAt;
   }

   @Override
   public String toString() {
      return new ToStringBuilder(this)
         .append("attributes", this.attributes)
         .append("isAgeOver", ToStringUtil.toString(this.ageOverList))
         .append("timeStamp", this.timeStamp != null ? DateFormatUtils.ISO_DATETIME_TIME_ZONE_FORMAT.format(this.timeStamp) : null)
         .append("validAt", this.validAt != null ? DateFormatUtils.ISO_DATETIME_TIME_ZONE_FORMAT.format(this.validAt) : null)
         .toString();
   }

   // SPECIFIC GETTER AND SETTER
   
   @Override
   public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result + ((this.ageOverList == null) ? 0 : this.ageOverList.hashCode());
      result = prime * result + ((this.attributes == null) ? 0 : this.attributes.hashCode());
      result = prime * result + ((this.timeStamp == null) ? 0 : this.timeStamp.hashCode());
      result = prime * result + ((this.validAt == null) ? 0 : this.validAt.hashCode());
      return result;
   }

   @Override
   public boolean equals(Object obj) {
      if (this == obj) {
         return true;
      }
      if (obj == null) {
         return false;
      }
      if (getClass() != obj.getClass()) {
         return false;
      }
      SubjectAttributes other = (SubjectAttributes) obj;
      if (this.ageOverList == null) {
         if (other.ageOverList != null) {
            return false;
         }
      } else if (!this.ageOverList.equals(other.ageOverList)) {
         return false;
      }
      if (this.attributes == null) {
         if (other.attributes != null) {
            return false;
         }
      } else if (!this.attributes.equals(other.attributes)) {
         return false;
      }
      if (this.timeStamp == null) {
         if (other.timeStamp != null) {
            return false;
         }
      } else if (!this.timeStamp.equals(other.timeStamp)) {
         return false;
      }
      if (this.validAt == null) {
         if (other.validAt != null) {
            return false;
         }
      } else if (!this.validAt.equals(other.validAt)) {
         return false;
      }
      return true;
   }

   // eIdentifier
   public String getEIdentifier() {
      return this.get(EIDENTIFIER);
   }
   
   public void setEIdentifier(String eIdentifier) {
      this.set(EIDENTIFIER, eIdentifier);
   }
   
   // date of birth
   
   public String getDateOfBirth() {
      return this.get(DATE_OF_BIRTH);
   }
   
   public void setDateOfBirth(String dateOfBirth) {
      this.set(DATE_OF_BIRTH, dateOfBirth);
   }
   
   public Date getDateOfBirthValue() throws EIDConnectorException {
      String dob = this.getDateOfBirth();
      if (dob == null) {
         return null;
      }

      // yyyy-MM-dd
      try {
         return DateUtils.parseDate(dob, new String[] { "yyyy-MM-dd" } );
      } catch (ParseException e) {
         throw new EIDConnectorException(ErrorCode.INVALID_DATE_FORMAT, e);
      }
      
   }
   
   public void setDateOfBirthValue(Date dateOfBirth) {
      String dob = DateFormatUtils.format(dateOfBirth, "yyyy-MM-dd");
      this.setDateOfBirth(dob);
   }
   
   // QAA level
   
   public String getQAALevel() {
      return this.get(CITIZEN_QAA_LEVEL);
   }
   
   public void setQAALevel(String qaaLevel) {
      this.set(CITIZEN_QAA_LEVEL, qaaLevel);
   }
   
   public Integer getQAALevelValue() throws EIDConnectorException {
      String qaa = this.getQAALevel();
      if (qaa == null) {
         return null;
      }
      try {
         return Integer.parseInt(qaa);
      } catch (NumberFormatException e) {
         throw new EIDConnectorException(ErrorCode.INVALID_QAA_LEVEL_FORMAT, e);
      }
   }
   
   public void setQAALevelValue(int qaaLevel) {
      this.set(CITIZEN_QAA_LEVEL, String.valueOf(qaaLevel));
   }
   
   // age
   
   public String getAge() {
      return this.get(AGE);
   }
   
   public void setAge(String age) {
      this.set(AGE, age);
   }
   
   public int getAgeValue() throws EIDConnectorException {
      String ageText = StringUtils.trimToNull(this.getAge());
      if (ageText == null) {
         throw new EIDConnectorException(ErrorCode.DATA_NOT_AVAILABLE, "Age attribute has not been set.");
      }
      try {
         return Integer.parseInt(ageText);
      } catch (NumberFormatException e) {
         throw new EIDConnectorException(ErrorCode.INVALID_AGE_FORMAT, e);
      }
   }
   
   public void setAgeValue(int age) {
      this.setAge(String.valueOf(age));
   }
   
   // given name
   
   public String getGivenName() {
      return this.get(GIVEN_NAME);
   }
   
   public void setGivenName(String givenName) {
      this.set(GIVEN_NAME, givenName);      
   }
   
   // age over
   public void addIsAgeOver(AgeOverData ageOverData) {
      this.ageOverList.add(ageOverData);
   }
   
   public AgeOverData[] getIsAgeOver() {
      return this.ageOverList.toArray(new AgeOverData[this.ageOverList.size()]);
   }
   
   // surname
   
   public String getSurName() {
      return this.get(SURNAME);
   }
   
   public void setSurName(String surName) {
      this.set(SURNAME, surName);      
   }

}
