/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.data;

import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import org.apache.log4j.Logger;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;

/**
 * Class that provides a store for {@link SubjectAttributes} resulting from successful
 * authentication. Putting SubjectAttributes in the store returns a ticket. For retrieval of the
 * stored attributes this ticket has to be provided. After retrieval the attributes are removed from
 * store invalidating the ticket.
 * @author AT-TUG (Thomas Knall)
 */
public final class SubjectAttributesStore {
   
   private final Logger log = Logger.getLogger(SubjectAttributesStore.class);
   private int keepTicketForSeconds;
   private boolean removeAfterUsage;

   private SubjectAttributesStore() {
      log.debug("Instantiating " + SubjectAttributesStore.class.getName() + ".");
      this.tickets = new Hashtable<String, SubjectAttributes>();
      this.keepTicketForSeconds = Configuration.getInternalIntProperty("ticket.validity.seconds");
      this.removeAfterUsage = Configuration.getInternalBooleanProperty("ticket.remove.after.usage");
      log.debug("Instantiating ticket store.");
      if (this.removeAfterUsage) {
         if (this.keepTicketForSeconds > 0) {
            log.debug("Tickets will be removed after each usage or automatically after " + this.keepTicketForSeconds + " seconds.");
         } else {
            log.warn("Tickets will be removed after each usage but NOT after a defined time period.");
         }
      } else {
         if (this.keepTicketForSeconds > 0) {
            log.warn("Tickets will NOT be removed after usage but automatically removed after " + this.keepTicketForSeconds + " seconds.");
         } else {
            log.warn("Tickets will NOT be removed, thus will be kept forever.");
         }
      }
   }
   
   private Hashtable<String, SubjectAttributes> tickets;
   
   private static final class LazyHolder {
      private static final SubjectAttributesStore SUBJECT_ATTRIBUTES_STORE_INSTANCE = new SubjectAttributesStore();
      private LazyHolder() {
      }
   }
   
   public static SubjectAttributesStore getInstance() {
      return LazyHolder.SUBJECT_ATTRIBUTES_STORE_INSTANCE;
   }
   
   public SubjectAttributes getSubjectAttributes(String ticket) {
      return this.retrieveSubjectAttributes(ticket, !this.removeAfterUsage);
   }
   
   SubjectAttributes retrieveSubjectAttributes(String ticket, boolean keep) {
      synchronized (this.tickets) {
         SubjectAttributes sa = keep ? this.tickets.get(ticket) : this.tickets.remove(ticket);
         long now = new Date().getTime();
         if (sa != null && this.keepTicketForSeconds > 0) {
            if (now - sa.getTimeStamp().getTime() > ((long) this.keepTicketForSeconds) * 1000) {
               log.debug("Ticket \"" + ticket + "\" found but removed since timestamp was too old.");
               this.tickets.remove(ticket);
            }
         }
         return sa;
      }
   }
   
   /**
    * Stores the {@link SubjectAttributes} from a successful authentication and returns a ticket.
    * @param attributes The SubjectAttributes.
    * @return The resulting ticket for later retrieval.
    * @throws NoSuchAlgorithmException Thrown if the underlying crypto/hash algorithms are
    * not available.
    */
   public String storeSubjectAttributes(SubjectAttributes attributes) throws NoSuchAlgorithmException {
      return this.storeSubjectAttributes(null, attributes);
   }
   
   /**
    * Stores the {@link SubjectAttributes} from a successful authentication and returns a ticket.
    * @param sourceID The sourceID.
    * @param attributes The SubjectAttributes.
    * @return The resulting ticket for later retrieval.
    * @throws NoSuchAlgorithmException Thrown if the underlying crypto/hash algorithms (SHA-1) are
    * not available.
    */
   public String storeSubjectAttributes(String sourceID, SubjectAttributes attributes) throws NoSuchAlgorithmException {
      String ticket = TicketProvider.getInstance().createTicket(sourceID);
      this.storeSubjectAttributesWithTimeStamp(ticket, attributes);
      return ticket;
   }
   
   void storeSubjectAttributesWithTimeStamp(String ticket, SubjectAttributes subjectAttributes) {
      subjectAttributes.setTimeStamp(new Date());
      synchronized (this.tickets) {
         this.tickets.put(ticket, subjectAttributes);
      }
   }
   
   /**
    * Invokes a cleanup method which removed tickets older than {@link #keepTicketForSeconds}
    * seconds. Note that {@link #keepTicketForSeconds} {@code <= 0} disables cleanup.
    */
   public void cleanUp() {
      if (this.keepTicketForSeconds > 0) {
         long now = new Date().getTime();
         synchronized (this.tickets) {
            Set<String> keys = new HashSet<String>(this.tickets.keySet());
            Iterator<String> keyIt = keys.iterator();
            while (keyIt.hasNext()) {
               String key = keyIt.next();
               SubjectAttributes sa = this.tickets.get(key);
               if (now - sa.getTimeStamp().getTime() > ((long) this.keepTicketForSeconds) * 1000) {
                  log.debug("CleanUp: Removing ticket \"" + key + "\".");
                  this.tickets.remove(key);
               }
            }
         }
      }
   }
   
}
