/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.data;

import iaik.security.random.SHA1Random;

import java.io.ByteArrayOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

import org.apache.log4j.Logger;

import at.iaik.commons.util.ConvertUtil;
import at.iaik.commons.util.MiscUtil;

/**
 * Provides a ticket creation service following OASIS SAML Binding 1.1.
 * @author AT-TUG (Thomas Knall)
 *
 */
public final class TicketProvider {

   /**
    * The underlying PRNG.
    */
   private static final Class< ? extends SecureRandom > SECURE_RANDOM_IMPL = SHA1Random.class;
   
   /**
    * The message digest algorithm for ticket creation.
    */
   private static final String MESSAGE_DIGEST_ALGORITHM = "SHA-1";
   
   private final Logger log = Logger.getLogger(TicketProvider.class);
   
   private SecureRandom random;
   private MessageDigest md;

   private static final class LazyHolder {
      private static final TicketProvider TICKET_PROVIDER_INSTANCE = new TicketProvider();
      private LazyHolder() {
      }
   }
   
   public static TicketProvider getInstance() {
      return LazyHolder.TICKET_PROVIDER_INSTANCE;
   }
   
   private TicketProvider() {
      log.debug("Instantiating " + TicketProvider.class.getName() + ".");
      try {
         this.random = SECURE_RANDOM_IMPL.newInstance();
      } catch (InstantiationException e) {
         log.error(e);
         log.warn("Using SecureRandom implementation from default security provider.");
         this.random = new SecureRandom();
      } catch (IllegalAccessException e) {
         log.error(e);
         log.warn("Using SecureRandom implementation from default security provider.");
         this.random = new SecureRandom();
      }
      try {
         this.md = MessageDigest.getInstance(MESSAGE_DIGEST_ALGORITHM);
         if (this.md.getDigestLength() == 0) {
            log.error("Unknown digest length of algorithm \"" + MESSAGE_DIGEST_ALGORITHM + "\". Instead of hashing data for the ticket creating, random data will be used.");
            this.md = null;
         }
      } catch (NoSuchAlgorithmException e) {
         log.error(e);
         log.warn("Instead of hashing data for the ticket creating, random data will be used.");
      }
   }
   
   public String createTicket() throws NoSuchAlgorithmException {
      return this.createTicket(null);
   }
   
   public String createTicket(String sourceID) throws NoSuchAlgorithmException {
      return this.createTicket(sourceID, null);
   }
   
   public String createTicket(String sourceID, String assertionHandle) throws NoSuchAlgorithmException {
      byte[] typeData = new byte[] { (byte)0x00, (byte)0x01 };
      byte[] sourceIDData = new byte[this.md != null ? this.md.getDigestLength() : 20];
      byte[] assertionHandleData = new byte[this.md != null ? this.md.getDigestLength() : 20];

      log.debug("Started ticket creation [0x0001 sourceID(" + sourceIDData.length + " bytes) assertionHandle(" + assertionHandleData.length + " bytes)].");
      if (MiscUtil.isEmpty(sourceID)) {
         log.debug("No sourceID provided. Filling sourceID with random data.");
         this.random.nextBytes(sourceIDData);
      } else if (this.md == null) {
         log.warn("Message digest algorithm \"" + MESSAGE_DIGEST_ALGORITHM + "\" is not available or does not provide data on digest length. Filling sourceID with random data.");
         this.random.nextBytes(sourceIDData);
      } else {
         log.debug("Using sourceID \"" + sourceID + "\".");
         this.md.reset();
         sourceIDData = this.md.digest(sourceID.getBytes());
      }

      if (MiscUtil.isEmpty(assertionHandle)) {
         log.debug("No assertionHandle provided. Filling assertionHandle with random data.");
         this.random.nextBytes(assertionHandleData);
      } else if (this.md == null) {
         log.warn("Message digest algorithm \"" + MESSAGE_DIGEST_ALGORITHM + "\" is not available or does not provide data on digest length. Filling assertionHandle with random data.");
         this.random.nextBytes(assertionHandleData);
      } else {
         log.debug("Using assertionHandle \"" + assertionHandle + "\".");
         this.md.reset();
         assertionHandleData = md.digest(assertionHandle.getBytes());
      }

      log.debug("Building ticket.");
      ByteArrayOutputStream out = new ByteArrayOutputStream(typeData.length + sourceIDData.length + assertionHandleData.length);
      out.write(typeData, 0, typeData.length);
      out.write(sourceIDData, 0, sourceIDData.length);
      out.write(assertionHandleData, 0, assertionHandleData.length);
      byte[] ticketData = out.toByteArray();
      String ticket = ConvertUtil.toBase64String(ticketData, true, false);
      log.debug("Resulting ticket: " + ticket);
      
      return ticket;
   }
   
}
