/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.impl;

import java.util.Hashtable;
import java.util.Set;

import org.apache.log4j.Logger;
import org.jconfig.Category;

import at.gv.egiz.moodle.eidconnector.auth.AuthenticationProvider;
import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.iaik.commons.util.MiscUtil;

/**
 * @author AT-TUG (Thomas Knall)
 */
public final class AuthenticationImplProvider {

   private final Logger log = Logger.getLogger(AuthenticationImplProvider.class);
   private Hashtable<String, AuthenticationProvider> authenticationProviderInstances; 

   private static final class LazyHolder {
      private static final AuthenticationImplProvider AUTHENTICATION_IMPL_PROVIDER_INSTANCE = new AuthenticationImplProvider();
      private LazyHolder() {
      }
   }
   
   public static AuthenticationImplProvider getInstance() {
      return LazyHolder.AUTHENTICATION_IMPL_PROVIDER_INSTANCE;
   }
   
   public String[] getAuthenticatedCountries() {
      synchronized (this.authenticationProviderInstances) {
         Set<String> keys = this.authenticationProviderInstances.keySet();
         String[] result = keys.toArray(new String[keys.size()]);
         
         // put stork authenticator on first position
         int i;
         for (i = 0; i < result.length && !Configuration.CONFIG_STORK_AUTHENTICATOR_ID.equalsIgnoreCase(result[i]); i++);
         if (i < result.length) {
            // stork found -> swap elements
            String tmp = result[0];
            result[0] = result[i];
            result[i] = tmp;
         }
         return result;
      }
   }
   
   public AuthenticationProvider getAuthenticationProvider(String countryCode) {
      synchronized (this.authenticationProviderInstances) {
         return this.authenticationProviderInstances.get(countryCode);
      }
   }
   
   private AuthenticationImplProvider() {
      log.debug("Instantiating " + AuthenticationImplProvider.class.getName() + ".");
      this.authenticationProviderInstances = new Hashtable<String, AuthenticationProvider>();
      Category[] cats = Configuration.getNestedCategories("auth");
      for (Category cat : cats) {
         String name = cat.getCategoryName();
         if (Configuration.isAuthCountryEnabled(name)) {
            String implClass = Configuration.getAuthCountryProperty(name, "authentication.provider.impl");
            if (MiscUtil.isEmpty(implClass)) {
               log.error("Authentication Provider implementation for country \"" + name + "\" is empty.");
            } else {
               try {
                  Object apObj = Class.forName(implClass).newInstance();
                  if (!(apObj instanceof AuthenticationProvider)) {
                     log.error("Declared Authentication Provider \"" + implClass + "\" for country \"" + name + "\" is not instance of " + AuthenticationProvider.class.getName() + ".");
                  } else {
                     AuthenticationProvider ap = (AuthenticationProvider) apObj;
                     log.info("Registering Authentication Provider [" + name + "]: " + implClass);
                     this.authenticationProviderInstances.put(name, ap);
                  }
               } catch (ClassNotFoundException e) {
                  log.error(e);
               } catch (InstantiationException e) {
                  log.error(e);
               } catch (IllegalAccessException e) {
                  log.error(e);
               }
            }
         }
      }
   }
   
}
