/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.impl;

import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.servlet.http.HttpServletRequest;

import org.apache.log4j.Logger;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.Attribute;
import org.opensaml.saml2.core.AttributeStatement;
import org.opensaml.saml2.core.StatusCode;
import org.opensaml.ws.transport.http.HttpServletRequestAdapter;
import org.opensaml.xml.util.XMLHelper;

import at.gv.egiz.moodle.eidconnector.EIDConnectorException;
import at.gv.egiz.moodle.eidconnector.ErrorCode;
import at.gv.egiz.moodle.eidconnector.auth.AuthenticationProvider;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributes;
import at.gv.egiz.moodle.eidconnector.logging.StatisticLoggerProvider;
import at.gv.egiz.moodle.eidconnector.validation.AssertionVerifier;
import at.gv.egiz.moodle.eidconnector.validation.PEPSConnectorAssertionVerifier;
import at.gv.egiz.moodle.eidconnector.validation.PEPSConnectorResponseVerifier;
import at.gv.egiz.moodle.eidconnector.validation.ResponseVerifier;
import at.iaik.commons.util.CryptUtil;
import at.iaik.commons.util.MiscUtil;
import eu.stork.mw.common.exceptions.SecurityException;
import eu.stork.mw.messages.saml.STORKResponse;
import eu.stork.vidp.messages.util.SAMLUtil;

public abstract class PEPSAuthenticationProvider implements AuthenticationProvider {

   protected static final String SAML_RESPONSE_PARAM_NAME = "SAMLResponse";

   private final Logger log = Logger.getLogger(PEPSAuthenticationProvider.class);

   public PEPSAuthenticationProvider() {
   }

   protected abstract STORKResponse getSTORKResponse(HttpServletRequest request) throws EIDConnectorException;

   public SubjectAttributes authenticate(HttpServletRequest request) throws EIDConnectorException {

      log.debug("Requested URL = " + request.getRequestURL());

      // check if post parameter is available
      if (MiscUtil.isEmpty(request.getParameter(SAML_RESPONSE_PARAM_NAME))) {
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Missing parameter \"" + SAML_RESPONSE_PARAM_NAME + "\"");
      }

      STORKResponse samlResponse = this.getSTORKResponse(request);
      log.debug("SAML response: " + XMLHelper.prettyPrintXML(samlResponse.getDOM()));

      // check status
      String statusCodeValue = samlResponse.getStatus().getStatusCode().getValue();
      if (!statusCodeValue.equals(StatusCode.SUCCESS_URI)) {
         log.error("Received ErrorResponse from PEPS: " + statusCodeValue);
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "ErrorResponse response from PEPS.");
      }
      log.debug("Got SAML response with success message.");


      // perform validation
      log.debug("Starting validation of SAML response");
      ResponseVerifier responseVerifier = new PEPSConnectorResponseVerifier();
      try {
         responseVerifier.verify(samlResponse);
      } catch (SecurityException e) {
         log.error("Error validating response message from PEPS.", e);
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Error validating response message from PEPS.");
      }

      log.info("Received SAML response from: " + samlResponse.getIssuer().getValue());
      log.info("SAML response succesfully verified!");



      // prepare retrieval of attributes
      if (samlResponse.getAssertions().isEmpty()) {
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Assertion missing in SAMLResponse.");
      }
      Assertion assertion = samlResponse.getAssertions().get(0);
      log.debug("Starting validation of SAML assertion");

      //validate Assertion
      AssertionVerifier assertionVerifier = new PEPSConnectorAssertionVerifier();
      try {
         // request id validation is disabled since we do not have a request id
         assertionVerifier.verify(assertion, new HttpServletRequestAdapter(request).getPeerAddress(), null, request.getRequestURL().toString());
      } catch (at.gv.egiz.moodle.eidconnector.validation.SecurityException e) {
         log.error("Error verifying assertion from PEPS", e);
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Error validating assertion received from PEPS.");
      }

      log.info("SAML assertion succesfully verified!");



      // retrieve attributes
      List<AttributeStatement> attributeStatements = assertion.getAttributeStatements();
      if (attributeStatements == null || attributeStatements.size() != 1) {
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "One single AttributeStatement element expected in SAMLResponse.");
      }
      AttributeStatement attributeStatement = attributeStatements.get(0);

      // iterate over all attributes and put them into eid connector store
      // attribute namespaces of eid connector store matches STORK attribute namespaces,
      // so we can directly push them into the store
      SubjectAttributes sa = new SubjectAttributes();
      List<Attribute> attributes = attributeStatement.getAttributes();
      if (attributes != null) {
         Iterator<Attribute> attributeIt = attributes.iterator();
         while (attributeIt.hasNext()) {
            Attribute attribute = attributeIt.next();
            String stringValue = SAMLUtil.getAttributeStringValue(attributes, attribute.getName());
            if (stringValue != null) {
               log.debug("Adding attribute \"" + attribute.getName() + "\".");
               sa.set(attribute.getName(), stringValue);
            }

         }
      }

      log.debug("PEPS Authentication Provider returns: " + sa);

      String issuer = null;
      if (assertion.getIssuer() != null) {
         issuer = assertion.getIssuer().getValue();
      }
      if (issuer == null && samlResponse.getIssuer() != null) {
         issuer = samlResponse.getIssuer().getValue();
      }

      // logging
      StatisticLoggerProvider.getInstance().log(
               new Date(),
               issuer,
               samlResponse.getDestination(),
               MiscUtil.isNotEmpty(sa.getEIdentifier()) ? CryptUtil.calcBase64Hash(sa.getEIdentifier().getBytes(), "SHA") : null,
               sa.get(SubjectAttributes.NATIONALITY),
               sa.getQAALevelValue(),
               sa.getAgeValue()
      );
      sa.setAlreadyLogged(false);

      return sa;
   }

   public boolean isResponsible(HttpServletRequest request) throws EIDConnectorException {
      return MiscUtil.isNotEmpty(request.getParameter(SAML_RESPONSE_PARAM_NAME));
   }

}
