/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.auth.impl.at;

import iaik.utils.Base64Exception;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Set;

import javax.net.ssl.SSLSocketFactory;
import javax.servlet.http.HttpServletRequest;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import asit.common.Utils;
import asit.moa.client.exception.ClientException;
import at.gv.egiz.moodle.eidconnector.EIDConnectorException;
import at.gv.egiz.moodle.eidconnector.ErrorCode;
import at.gv.egiz.moodle.eidconnector.auth.AuthenticationProvider;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributes;
import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.logging.StatisticLoggerProvider;
import at.gv.egiz.moodle.eidconnector.webapp.struts.Constants;
import at.iaik.commons.data.ssl.KeyStore;
import at.iaik.commons.moa.ClientFactory;
import at.iaik.commons.moa.id.generic.GenericMOAIDClient;
import at.iaik.commons.util.CryptUtil;
import at.iaik.commons.util.DateTimeUtil;
import at.iaik.commons.util.IOUtil;
import at.iaik.commons.util.MiscUtil;
import at.iaik.commons.util.WebAppUtil;
import at.iaik.commons.xmlschema.saml.AgeOverData;
import at.iaik.commons.xmlschema.saml.SAMLAssertion;

/**
 * @author AT-TUG (Thomas Knall)
 */
public class AustrianAuthProviderImpl implements AuthenticationProvider {
   
   private final Logger log = LoggerFactory.getLogger(AustrianAuthProviderImpl.class);
   private static final String SAML_ARTIFACT_REQUEST_PARAMETER = "SAMLArtifact";
   
   private static Set<String> alreadyLogged = new HashSet<String>();
   private static Hashtable<String, SSLSocketFactory> sslSocketFactoryInstances = new Hashtable<String, SSLSocketFactory>();
   
   public GenericMOAIDClient getMOAIDClient(String instance) {
      GenericMOAIDClient moaidclient = null;
      try {
         log.debug("Instantiating MOAIDFilterCient.");
         moaidclient = ClientFactory.getGenericMOAIDClient();
         String getAuthDataURL = Configuration.getMOAIDProperty(instance, "getauthenticationdata.url");
         moaidclient.setDefaultAddress(getAuthDataURL);
         log.debug("Setting authDataURL = " + getAuthDataURL);
         if ("https".equalsIgnoreCase(new URL(getAuthDataURL).getProtocol())) {
            log.debug("SSL connection detected.");
            SSLSocketFactory sslsocketfactory = this.getSSLSocketFactory(instance);
            if (sslsocketfactory != null) {
               moaidclient.setSSLSocketFactory(sslsocketfactory);
            }
         }
      } catch (Exception e) {
         log.error("Error instantiating MOA-ID client.", e);
      }
      return moaidclient;
   }

   public KeyStore getClientKeyStore(String instance) throws IOException {
      KeyStore clientKeyStore = new KeyStore();
      byte[] keyStoreData = IOUtil.toByteArray(IOUtil.getResourceAsStream(Configuration.getMOAIDProperty(instance, "keystore.uri")));
      clientKeyStore.setData(keyStoreData);
      clientKeyStore.setPassword(Configuration.getMOAIDProperty(instance, "keystore.password"));
      clientKeyStore.setType(Configuration.getMOAIDProperty(instance, "keystore.type"));
      log.debug("Returning client key store.");
      return clientKeyStore;
   }

   public KeyStore getTrustKeyStore(String instance) throws IOException {
      KeyStore trustKeyStore = new KeyStore();
      byte[] trustStoreData = IOUtil.toByteArray(IOUtil.getResourceAsStream(Configuration.getMOAIDProperty(instance, "truststore.uri")));
      trustKeyStore.setData(trustStoreData);
      trustKeyStore.setPassword(Configuration.getMOAIDProperty(instance, "truststore.password"));
      trustKeyStore.setType(Configuration.getMOAIDProperty(instance, "truststore.type"));
      log.debug("Returning trust key store.");
      return trustKeyStore;
   }
   
   protected synchronized SSLSocketFactory getSSLSocketFactory(String instance) throws Exception {
      if (sslSocketFactoryInstances.get(instance) == null) {
         KeyStore clientKeyStore = this.getClientKeyStore(instance);
         KeyStore trustKeyStore = this.getTrustKeyStore(instance);
         
         SSLSocketFactory sslSocketFactory = null;
         if (clientKeyStore != null && clientKeyStore.isValid() && trustKeyStore != null && trustKeyStore.isValid()) {
            log.debug("Creating SSLSocketFactory for instance \"" + instance + "\".");
            sslSocketFactory = Utils.configureHTTPSConnection(
                  new ByteArrayInputStream(clientKeyStore.getData()), clientKeyStore.getPassword(), clientKeyStore.getType(),
                  new ByteArrayInputStream(trustKeyStore.getData()), trustKeyStore.getPassword(), trustKeyStore.getType()
            );
         } else {
            if (!alreadyLogged.contains(instance)) {
               log.warn("In order to fetch authentication data via ssl both client key store and trusted key store have to be provided for instance \"" + instance + "\". Ignore this message if authentication data is retrieved via non ssl.");
               alreadyLogged.add(instance);
            }
            return null;
         }
         sslSocketFactoryInstances.put(instance, sslSocketFactory);
      }
      
      return sslSocketFactoryInstances.get(instance);
   }

   public SubjectAttributes authenticate(HttpServletRequest request) throws EIDConnectorException {
      String samlArtifact = request.getParameter(SAML_ARTIFACT_REQUEST_PARAMETER);
      if (MiscUtil.isEmpty(samlArtifact)) {
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Missing request parameter \"" + SAML_ARTIFACT_REQUEST_PARAMETER + "\"");
      }
      
      // iterator over all possible MOA-ID instances
      String[] instances = Configuration.getMOAIDInstanceNames();
      String originatingInstance = Constants.DEFAULT_STORK_MOA_ID_CATEGORY_NAME;
      for (String instance : instances) {
         log.debug("Evaluating instance \"" + instance + "\" for responsibility.");
         try {
            if (WebAppUtil.isArtifactFromSourceID(samlArtifact, instance)) {
               originatingInstance = instance;
               break;
            }
         } catch (Base64Exception e) {
            throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Error decoding SAML artifact \"" + samlArtifact + "\".");
         } catch (NoSuchAlgorithmException e) {
            throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Error instanciating digest algorithm.", e);
         }
      }
      log.debug("SAML artifact \"" + samlArtifact + "\" originates from instance \"" + originatingInstance + "\".");
      SAMLAssertion samlAssertion;
      try {
         GenericMOAIDClient client = this.getMOAIDClient(originatingInstance);
         client.setSAMLArtifact(samlArtifact);
         samlAssertion = client.getAuthenticationData();
      } catch (ClientException c) {
         throw new EIDConnectorException(ErrorCode.AUTH_PROVIDER_ERROR, "Error retrieving authentication data (" + c.getMessage() + ")", c);
      }
      
      SubjectAttributes sa = new SubjectAttributes();

      // age
      if (!Configuration.isForceAgeCalculationEnabled()) {
         if (samlAssertion.getAge() != null) {
            sa.setAgeValue(samlAssertion.getAge());
         }
      } else {
         log.warn("Age is going to be calculated from date of birth instead of using the respective STORK subject attribute.");
      }
      
      // set date of birth (mandatory)
      if (samlAssertion.getDateOfBirth() != null) {
         sa.setDateOfBirthValue(samlAssertion.getDateOfBirth());
         if (samlAssertion.getDateOfIssue() != null) {
            sa.setValidAt(samlAssertion.getDateOfIssue());
            if (sa.getAge() == null) {
               sa.setAgeValue(DateTimeUtil.calcAge(samlAssertion.getDateOfBirth(), samlAssertion.getDateOfIssue()));
            }
         }
      }
      
      // isAgeOver
      if (samlAssertion.getIsAgeOver() != null) {
         AgeOverData[] ageOverData = samlAssertion.getIsAgeOver();
         for (AgeOverData ageOver : ageOverData) {
            sa.addIsAgeOver(ageOver);
         }
      }
      
      // nationality
      if (samlAssertion.getNationalityCode() != null) {
         sa.set(SubjectAttributes.NATIONALITY, String.valueOf(samlAssertion.getNationalityCode()));
      } else {
         String nationalityCode = Configuration.getMOAIDNationalityCode(originatingInstance);
         if (nationalityCode != null) {
            log.debug("Nationality code not given by MOA, assuming default value for instance \"{}\": {}", originatingInstance, nationalityCode);
            sa.set(SubjectAttributes.NATIONALITY, nationalityCode);
         }
      }
      
      // Citizen QAA level
      if (samlAssertion.getCitizenQAAlevel() != null) {
         sa.set(SubjectAttributes.CITIZEN_QAA_LEVEL, String.valueOf(samlAssertion.getCitizenQAAlevel()));
      } else {
         Integer level = Configuration.getMOAIDQAALevel(originatingInstance);
         if (level != null) {
            log.debug("Citizen QAA level not given by MOA, assuming default value for instance \"" + originatingInstance + "\": " + level);
            sa.set(SubjectAttributes.CITIZEN_QAA_LEVEL, String.valueOf(level));
         }
      }
      
      // title
      if (samlAssertion.getTitle() != null) {
         sa.set(SubjectAttributes.TITLE, samlAssertion.getTitle());
      }
      
      // pseudonym
      if (samlAssertion.getPseudonym() != null) {
         sa.set(SubjectAttributes.PSEUDONYM, samlAssertion.getPseudonym());
      }
      
      // set eIdentifier (mandatory)
      if (MiscUtil.isNotEmpty(samlAssertion.getNameIdentifier())) {
         sa.setEIdentifier(samlAssertion.getNameIdentifier());
      }
      
      // set eQualifier (optional)
      if (MiscUtil.isNotEmpty(samlAssertion.getNameQualifier())) {
         sa.set(SubjectAttributes.EQUALIFIER, samlAssertion.getNameQualifier());
      }
      
      // set given name if available
      if (MiscUtil.isNotEmpty(samlAssertion.getGivenName())) {
         sa.setGivenName(samlAssertion.getGivenName());
      }
      
      // set family name if available
      if (MiscUtil.isNotEmpty(samlAssertion.getFamilyName())) {
         sa.setSurName(samlAssertion.getFamilyName());
      }
      
      // logging
      StatisticLoggerProvider.getInstance().log(
               new Date(),
               originatingInstance,
               Configuration.getMoodleProperty("moodle.ticket.consumer.service"),
               MiscUtil.isNotEmpty(sa.getEIdentifier()) ? CryptUtil.calcBase64Hash(sa.getEIdentifier().getBytes(), "SHA") : null,
               sa.get(SubjectAttributes.NATIONALITY),
               sa.getQAALevelValue(),
               sa.getAgeValue(),
               samlAssertion.getBkuURL(),
               samlAssertion.getNameQualifier()
      );         
      sa.setAlreadyLogged(true);
      
      return sa;
   }

   public boolean isResponsible(HttpServletRequest request) throws EIDConnectorException {
      return MiscUtil.isNotEmpty(request.getParameter(SAML_ARTIFACT_REQUEST_PARAMETER));
   }

}
