/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.core.config;

import static org.apache.log4j.Level.ALL;
import static org.apache.log4j.Level.DEBUG;
import static org.apache.log4j.Level.ERROR;
import static org.apache.log4j.Level.FATAL;
import static org.apache.log4j.Level.INFO;
import static org.apache.log4j.Level.WARN;

import java.net.URL;

import org.apache.log4j.Logger;
import org.jconfig.Category;

import at.iaik.commons.jconfig.JConfigDefaultConfiguration;
import at.iaik.commons.jconfig.JConfigDefaultValidator;
import at.iaik.commons.util.JConfigUtil;
import at.iaik.commons.util.MiscUtil;

/**
 * @author AT-TUG (Thomas Knall)
 */
public class JConfigValidatorImpl extends JConfigDefaultValidator {

   private Logger log = Logger.getLogger(getClass());

   public JConfigValidatorImpl(JConfigDefaultConfiguration cfg) {
		super(cfg);
	}

	@Override
	public void doValidate() {
		
		assertConfigurationWasLoaded(FATAL);

      assertMoodle();
      assertGeneral();
      assertInternal();
      assertHibernate(Configuration.HBM_NAME_EID_CONNECTOR);
      assertHibernate(Configuration.HBM_NAME_MOODLE);
      assertAuth();
      assertEIDHTTPHeaderMappings();
		assertError();
		assertSTORKResponseValidation();
		assertMOAID();
		
	}
	
   private void assertHibernate(String category) {
      if (assertCategory("hibernate/" + category, FATAL)) {
         assertProperty("hibernate.bytecode.use_reflection_optimizer", INFO);
         if (assertProperty("hibernate.hbm2ddl.auto", DEBUG)) {
            assertPropertyFormat("hibernate.hbm2ddl.auto", new String[] { "none", "validate", "update", "create", "create-drop" });
         }
         if (assertProperty("hibernate.cache.provider_class", INFO)) {
            assertProperty("net.sf.ehcache.configurationResourceName", INFO);
         }
         if (assertProperty("hibernate.connection.provider_class", WARN)) {
            assertIntProperty("hibernate.c3p0.acquire_increment", INFO);
            assertIntProperty("hibernate.c3p0.idle_test_period", INFO);
            assertIntProperty("hibernate.c3p0.timeout", INFO);
            assertIntProperty("hibernate.c3p0.max_size", INFO);
            assertIntProperty("hibernate.c3p0.max_statements", INFO);
            assertIntProperty("hibernate.c3p0.min_size", INFO);
            assertBooleanProperty("hibernate.c3p0.testConnectionOnCheckin", INFO);
            assertBooleanProperty("hibernate.c3p0.testConnectionOnCheckout", INFO);
            assertProperty("hibernate.c3p0.preferredTestQuery", INFO);
         }
         assertBooleanProperty("hibernate.connection.autoReconnect", INFO);
         assertBooleanProperty("hibernate.connection.autoReconnectForPools", INFO);
         assertBooleanProperty("hibernate.connection.is-connection-validation-required", INFO);
         assertProperty("hibernate.current_session_context_class", WARN);
         assertProperty("hibernate.dialect", ERROR);
         assertBooleanProperty("hibernate.connection.useUnicode", WARN);
         assertProperty("hibernate.connection.characterEncoding", WARN);
         assertProperty("hibernate.connection.charSet", WARN);
         if (assertProperty("hibernate.show_sql", DEBUG)) {
            assertProperty("hibernate.format_sql", DEBUG);
         }

         assertProperty("hibernate.connection.driver_class", ERROR);
         assertProperty("hibernate.connection.url", ERROR);
         assertProperty("hibernate.connection.username", INFO);
         assertProperty("hibernate.connection.password", INFO);
      }
   }
   
   private void assertMOAID() {
      if (assertCategory("moaid", INFO)) {

         JConfigDefaultConfiguration cfg = super.getJConfigDefaultConfiguration();
         Category cat = cfg.getCategory("moaid");
         Category[] nc = JConfigUtil.getNestedCategories(cat);
         for (Category lCat : nc) {
            assertMOAIDInstance(lCat.getCategoryName());
         }

      }
   }
   
   private void assertMOAIDInstance(String instance) {
      if (assertCategory("moaid/" + instance, ERROR)) {
         if (assertURLProperty("getauthenticationdata.url")) {
            JConfigDefaultConfiguration cfg = super.getJConfigDefaultConfiguration();
            URL url = cfg.getURLProperty("getauthenticationdata.url", "moaid/" + instance);
            if ("https".equalsIgnoreCase(url.getProtocol())) {
               assertProperty("keystore.uri");
               assertProperty("keystore.type");
               assertProperty("keystore.password", WARN);
               assertProperty("key.password", WARN);
               assertProperty("truststore.uri");
               assertProperty("truststore.type");
               assertProperty("truststore.password", WARN);
            }
            if (assertIntProperty("default.qaa.level", DEBUG)) {
               assertIntRangeProperty("default.qaa.level", ERROR, 1, 4);
            }
            if (assertProperty("default.nationality.code", DEBUG)) {
               assertCountryCodeProperty("default.nationality.code", ERROR);
            }
         }
      }
   }
   
	private void assertGeneral() {
	   if (assertCategory("general", FATAL)) {
	      assertProperty("resourcebundle");
	   }
	}
	
	private void assertInternal() {
      if (assertCategory("internal", ERROR)) {
         assertIntRangeProperty("ticket.validity.seconds", 0, Integer.MAX_VALUE);
         assertBooleanProperty("ticket.remove.after.usage");
         assertBooleanProperty("show.authenticator.errors");
         assertBooleanProperty("maintenancemode", WARN);
         assertProperty("eid-connector.httpheader.service", WARN);
         assertProperty("eid-connector.httpheader.service.value", WARN);
         assertBooleanProperty("force.calculate.age", DEBUG);
         assertBooleanProperty("eid.transformation.enabled", WARN);
         assertProperty("logger.impl", DEBUG);
         assertBooleanProperty("logger.database.enabled", WARN);
      }
	}
	
	private void assertError() {
		if (assertCategory("error", ERROR)) {
			assertProperty("mailto", WARN);
			assertProperty("cc", INFO);
			assertProperty("bcc", INFO);
			assertProperty("subject", INFO);
			assertProperty("body", INFO);
		}
	}
	
	private void assertEIDHTTPHeaderMappings() {
      // eid-http-header-mappings
      if (assertCategory("eid-http-header-mappings", INFO)) {
         JConfigDefaultConfiguration cfg = super.getJConfigDefaultConfiguration();
         Category cat = cfg.getCategory("eid-http-header-mappings");
         Category[] nc = JConfigUtil.getNestedCategories(cat);
         for (Category lCat : nc) {
            assertProperty("eid-http-header-mappings/" + lCat.getCategoryName(), "eid-attribute", ERROR);
            assertProperty("eid-http-header-mappings/" + lCat.getCategoryName(), "http-header", ERROR);
         }
      }
	}
	
	private void assertAuth() {
      if (assertCategory("auth", INFO)) {
         JConfigDefaultConfiguration cfg = super.getJConfigDefaultConfiguration();
         Category cat = cfg.getCategory("auth");
         Category[] nc = JConfigUtil.getNestedCategories(cat);
         for (Category lCat : nc) {
            if (!Configuration.CONFIG_STORK_AUTHENTICATOR_ID.equalsIgnoreCase(lCat.getCategoryName()) && !MiscUtil.isValidCountryCode(lCat.getCategoryName(), false)) {
               log.error("Country code (ISO 3166-1 alpha-2) category name \"" + lCat.getCategoryName() + "\" found but invalid in category \"auth\" of configuration.");
               updateMaxSeverity(ERROR);
            }
            assertProperty("auth/" + lCat.getCategoryName(), "startauthentication.url", INFO);
            assertProperty("auth/" + lCat.getCategoryName(), "authentication.provider.impl", ERROR);
            assertBooleanProperty("auth/" + lCat.getCategoryName(), "enabled", INFO);
         }
      }
	}
   
	private void assertMoodle() {
	   if (assertCategory("moodle", ERROR)) {
	      assertProperty("moodle.ticket.consumer.service", ERROR);
	      assertProperty("moodle.start.page", ERROR);
	      assertProperty("moodle.httpheader.username", ERROR);
	      assertProperty("moodle.allow.auth.types", ERROR);
	      assertMessageDigestProperty("moodle.hash.algorithm.password", ERROR);
	      assertMessageDigestProperty("moodle.hash.algorithm.username", ERROR);
	      assertBooleanProperty("moodle.existing.accounts", ERROR);
	      assertProperty("moodle.password.salt", INFO);
	      assertCharsetProperty("moodle.password.oldcharset", WARN);
	      assertBooleanProperty("moodle.new.accounts", ERROR);
	      if (assertBooleanProperty("moodle.existing.accounts", ALL) && assertBooleanProperty("moodle.new.accounts", ALL)) {
	         boolean newAccounts = super.getJConfigDefaultConfiguration().getBooleanProperty("moodle.new.accounts", true, "moodle");
	         boolean existingAccounts = super.getJConfigDefaultConfiguration().getBooleanProperty("moodle.existing.accounts", true, "moodle");
	         if (!newAccounts && !existingAccounts) {
	            log.error("At least on of \"moodle.existing.accounts\" and \"moodle.new.accounts\" has to be set in category \"moodle\".");
	            updateMaxSeverity(ERROR);
	         }
	      }
	   }
	}
	
   private void assertSTORKResponseValidation() {
      if (assertCategory("STORKResponseValidation", ERROR)) {
         assertProperty("trusted.issuers", ERROR);
         assertProperty("expected.audience", ERROR);
         assertProperty("signature-verification.config-path", ERROR);
         assertProperty("signature-verification.profile", ERROR);
      }
   }

	@Override
	public String getApplicationName() {
		return null;
	}

}
