/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.data;

import java.io.Serializable;
import java.util.Date;

import javax.persistence.Basic;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.Table;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.lang.time.DateFormatUtils;

/**
 * @author AT-TUG (Thomas Knall)
 */
@Entity
@Table(name = "statistic_log")
public class StatisticLog implements Serializable {

   private final static String TIMESTAMP_PATTERN = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
   
   private static final long serialVersionUID = 1L;
   
   
   // fields
   
   @Id @GeneratedValue(strategy = GenerationType.AUTO)
   private Integer id;

   @Temporal(TemporalType.TIMESTAMP)
   @Column(name = "timestamp", nullable = true, unique = false)
   private Date timestamp;
   
   @Basic
   @Column(name = "assertion_issuer_id", nullable = true, unique = false)
   private String assertionIssuerId;
   
   @Basic
   @Column(name = "service_provider_id", nullable = true, unique = false)
   private String serviceProviderId;
   
   @Basic
   @Column(name = "citizen_eid_digest", nullable = true, unique = false)
   private String citizenEIDDigest;
   
   @Basic
   @Column(name = "citizen_country", nullable = true, unique = false)
   private String citizenCountry;
   
   @Basic
   @Column(name = "citizen_qaa_level", nullable = true, unique = false)
   private Integer citizenQAALevel;
   
   @Basic
   @Column(name = "citizen_age", nullable = true, unique = false)
   private Integer citizenAge;
   
   @Basic
   @Column(name = "custom_field_1", nullable = true, unique = false)
   private String customField1;
   
   @Basic
   @Column(name = "custom_field_2", nullable = true, unique = false)
   private String customField2;
   
   // constructor
   public StatisticLog() {
      this.id = null;
      this.timestamp = null;
      this.assertionIssuerId = null;
      this.serviceProviderId = null;
      this.citizenEIDDigest = null;
      this.citizenCountry = null;
      this.citizenQAALevel = null;
      this.citizenAge = null;
      this.customField1 = null;
      this.customField2 = null;
   }

   /**
    * Creates a statistical log entry using the data provided.
    * Note that timestamp, assertionIssuerId and spID must not be {@code null}, while the remaining parameters
    * might be.<br/>
    * @param timestamp The current date.
    * @param assertionIssuerId The identifier of the underlying S-PEPS/V-IDP instance.
    * @param spID The identifier of the service provider being served.
    * @param sha1OfEID A base64 encoded String of the SHA1 hash of the citizen's eIdentifier.
    * @param citizenCountryCode The citizen's home country code (according to ISO 3166-1 alpha-2).
    * @param citizenQAALevel The underlying QAA level.
    * @param age The citizen's age.
    * @param customField1 Any custom data (or {@code null}) to be logged.
    * @param customField2 Any custom data (or {@code null}) to be logged.
    */
   public StatisticLog(Date timestamp, String assertionIssuerId, String spID, String sha1OfEID, String citizenCountryCode, Integer citizenQAALevel, Integer age, String customField1, String customField2) {
      this();
      this.timestamp = timestamp;
      this.assertionIssuerId = assertionIssuerId;
      this.serviceProviderId = spID;
      this.citizenEIDDigest = sha1OfEID;
      this.citizenCountry = citizenCountryCode;
      this.citizenQAALevel = citizenQAALevel;
      this.citizenAge = age;
      this.customField1 = customField1;
      this.customField2 = customField2;
   }
   
   /**
    * Creates a statistical log entry using the data provided.
    * Note that timestamp, assertionIssuerId and spID must not be {@code null}, while the remaining parameters
    * might be.<br/>
    * @param timestamp The current date.
    * @param assertionIssuerId The identifier of the underlying S-PEPS/V-IDP instance.
    * @param spID The identifier of the service provider being served.
    * @param sha1OfEID A base64 encoded String of the SHA1 hash of the citizen's eIdentifier.
    * @param citizenCountryCode The citizen's home country code (according to ISO 3166-1 alpha-2).
    * @param citizenQAALevel The underlying QAA level.
    * @param age The citizen's age.
    */
   public StatisticLog(Date timestamp, String assertionIssuerId, String spID, String sha1OfEID, String citizenCountryCode, Integer citizenQAALevel, Integer age) {
      this(timestamp, assertionIssuerId, spID, sha1OfEID, citizenCountryCode, citizenQAALevel, age, null, null);
   }
   
   // getter/setter
   
   public Integer getId() {
      return this.id;
   }

   public void setId(Integer id) {
      this.id = id;
   }

   public Date getTimestamp() {
      return this.timestamp;
   }

   public void setTimestamp(Date timestamp) {
      this.timestamp = timestamp;
   }

   public String getAssertionIssuerId() {
      return this.assertionIssuerId;
   }

   public void setAssertionIssuerId(String assertionIssuerId) {
      this.assertionIssuerId = assertionIssuerId;
   }

   public String getServiceProviderId() {
      return this.serviceProviderId;
   }

   public void setServiceProviderId(String serviceProviderId) {
      this.serviceProviderId = serviceProviderId;
   }

   public String getCitizenEIDDigest() {
      return this.citizenEIDDigest;
   }

   public void setCitizenEIDDigest(String citizenEIDDigest) {
      this.citizenEIDDigest = citizenEIDDigest;
   }

   public String getCitizenCountry() {
      return this.citizenCountry;
   }

   public void setCitizenCountry(String citizenCountry) {
      this.citizenCountry = citizenCountry;
   }

   public Integer getCitizenQAALevel() {
      return this.citizenQAALevel;
   }

   public void setCitizenQAALevel(Integer citizenQAALevel) {
      this.citizenQAALevel = citizenQAALevel;
   }

   public Integer getCitizenAge() {
      return this.citizenAge;
   }

   public void setCitizenAge(Integer citizenAge) {
      this.citizenAge = citizenAge;
   }

   public String getCustomField1() {
      return this.customField1;
   }

   public void setCustomField1(String customField1) {
      this.customField1 = customField1;
   }

   public String getCustomField2() {
      return this.customField2;
   }

   public void setCustomField2(String customField2) {
      this.customField2 = customField2;
   }

   @Override
   public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result
               + ((this.assertionIssuerId == null) ? 0 : this.assertionIssuerId.hashCode());
      result = prime * result + ((this.citizenAge == null) ? 0 : this.citizenAge.hashCode());
      result = prime * result
               + ((this.citizenCountry == null) ? 0 : this.citizenCountry.hashCode());
      result = prime * result
               + ((this.citizenEIDDigest == null) ? 0 : this.citizenEIDDigest.hashCode());
      result = prime * result
               + ((this.citizenQAALevel == null) ? 0 : this.citizenQAALevel.hashCode());
      result = prime * result + ((this.customField1 == null) ? 0 : this.customField1.hashCode());
      result = prime * result + ((this.customField2 == null) ? 0 : this.customField2.hashCode());
      result = prime * result + ((this.id == null) ? 0 : this.id.hashCode());
      result = prime * result
               + ((this.serviceProviderId == null) ? 0 : this.serviceProviderId.hashCode());
      result = prime * result + ((this.timestamp == null) ? 0 : this.timestamp.hashCode());
      return result;
   }

   @Override
   public boolean equals(Object obj) {
      if (this == obj)
         return true;
      if (obj == null)
         return false;
      if (getClass() != obj.getClass())
         return false;
      StatisticLog other = (StatisticLog) obj;
      if (this.assertionIssuerId == null) {
         if (other.assertionIssuerId != null)
            return false;
      } else if (!this.assertionIssuerId.equals(other.assertionIssuerId))
         return false;
      if (this.citizenAge == null) {
         if (other.citizenAge != null)
            return false;
      } else if (!this.citizenAge.equals(other.citizenAge))
         return false;
      if (this.citizenCountry == null) {
         if (other.citizenCountry != null)
            return false;
      } else if (!this.citizenCountry.equals(other.citizenCountry))
         return false;
      if (this.citizenEIDDigest == null) {
         if (other.citizenEIDDigest != null)
            return false;
      } else if (!this.citizenEIDDigest.equals(other.citizenEIDDigest))
         return false;
      if (this.citizenQAALevel == null) {
         if (other.citizenQAALevel != null)
            return false;
      } else if (!this.citizenQAALevel.equals(other.citizenQAALevel))
         return false;
      if (this.customField1 == null) {
         if (other.customField1 != null)
            return false;
      } else if (!this.customField1.equals(other.customField1))
         return false;
      if (this.customField2 == null) {
         if (other.customField2 != null)
            return false;
      } else if (!this.customField2.equals(other.customField2))
         return false;
      if (this.id == null) {
         if (other.id != null)
            return false;
      } else if (!this.id.equals(other.id))
         return false;
      if (this.serviceProviderId == null) {
         if (other.serviceProviderId != null)
            return false;
      } else if (!this.serviceProviderId.equals(other.serviceProviderId))
         return false;
      if (this.timestamp == null) {
         if (other.timestamp != null)
            return false;
      } else if (!this.timestamp.equals(other.timestamp))
         return false;
      return true;
   }

   @Override
   public String toString() {
      return new ToStringBuilder(this)
         .append("id", this.id)
         .append("timestamp", this.timestamp != null ? DateFormatUtils.formatUTC(timestamp, TIMESTAMP_PATTERN) : null)
         .append("assertionIssuerId", this.assertionIssuerId)
         .append("serviceProviderId", this.serviceProviderId)
         .append("citizenEIDDigest", this.citizenEIDDigest)
         .append("citizenCountry", this.citizenCountry)
         .append("citizenQAALevel", this.citizenQAALevel)
         .append("citizenAge", this.citizenAge)
         .append("customField1", this.customField1)
         .append("customField2", this.customField2)
      .toString();
   }

}
