/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.data.ap;

import java.util.Date;

import org.apache.commons.lang.time.StopWatch;
import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.data.IdMapping;
import at.iaik.commons.util.HibernateUtil;
import at.iaik.commons.util.MiscUtil;
import at.iaik.commons.util.ToStringUtil;

public class IdMappingAP {

   private static Logger LOG = LoggerFactory.getLogger(IdMappingAP.class);

   protected static IdMapping findIdMappingByCriterion(Criterion... criterions) {
      MiscUtil.assertNotEmpty(criterions, "Criterions");
      Transaction tx = null;
      IdMapping result = null;
      StopWatch sw = new StopWatch();
      sw.start();
      try {
         Session session = HibernateUtil.getCurrentSession(Configuration.HBM_NAME_EID_CONNECTOR);
         tx = session.beginTransaction();

         Criteria criteria = session.createCriteria(IdMapping.class);
         if (MiscUtil.isNotEmpty(criterions)) {
            LOG.debug("Looking for id mapping (criterion(s) = " + ToStringUtil.toString(criterions) + ").");
            for (Criterion criterion : criterions) {
               criteria.add(criterion);
            }
         }
         LOG.debug("Effective criteria: " + criteria.toString());
         result = (IdMapping) criteria.uniqueResult();
         tx.commit();
         sw.stop();
         LOG.debug((result != null ? "Found id mapping." : "Unable to find any id mapping.") + " Query time: " + sw.getTime() + "ms.");
      } catch (HibernateException hex) {
         LOG.error("Error searching for id mapping.", hex);
         if (tx != null) {
            LOG.debug("Initiating rollback.");
            try {
               tx.rollback();
            } catch (HibernateException rollbackFailed) {
               LOG.error("Rollback failed (reason: " + rollbackFailed.getMessage() + ")");
            }
         }
         sw.stop();
         LOG.debug("Transaction aborted after " + sw.getTime() + "ms.");
         throw hex;
      }
      return result;
   }

   public static IdMapping findIdMappingByNameIdentifier(String nameIdentifier) {
      MiscUtil.assertNotNull(nameIdentifier, "nameIdentifier");
      return findIdMappingByCriterion(Restrictions.eq("nameIdentifier", nameIdentifier));
   }

   public static IdMapping findIdMappingByPersonData(String givenNames, String familyName, Date dateOfBirth) {
      MiscUtil.assertNotNull(givenNames, "givenNames");
      MiscUtil.assertNotNull(familyName, "familyName");
      MiscUtil.assertNotNull(dateOfBirth, "dateOfBirth");
      return findIdMappingByCriterion(
               Restrictions.and(
                        Restrictions.and(
                                 Restrictions.eq("givenNames", givenNames),
                                 Restrictions.eq("familyName", familyName)
                        ),
                        Restrictions.eq("dateOfBirth", dateOfBirth)
               )
      );
   }

   public static IdMapping findIdMappingByPersonDataAndIdentifier(String givenNames, String familyName, Date dateOfBirth, String nameIdentifier) {
      MiscUtil.assertNotNull(givenNames, "givenNames");
      MiscUtil.assertNotNull(familyName, "familyName");
      MiscUtil.assertNotNull(dateOfBirth, "dateOfBirth");
      MiscUtil.assertNotNull(nameIdentifier, "nameIdentifier");
      return findIdMappingByCriterion(
               Restrictions.and(
                        Restrictions.and(
                                 Restrictions.eq("givenNames", givenNames),
                                 Restrictions.eq("familyName", familyName)
                        ),
                        Restrictions.and(
                                 Restrictions.eq("dateOfBirth", dateOfBirth),
                                 Restrictions.eq("nameIdentifier", nameIdentifier)
                        )
               )
             );
   }

}
