/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.data.ap;

import static at.gv.egiz.moodle.eidconnector.core.config.Configuration.HBM_NAME_MOODLE;
import iaik.utils.Util;

import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.lang.time.StopWatch;
import org.apache.log4j.Logger;
import org.hibernate.Criteria;
import org.hibernate.HibernateException;
import org.hibernate.Session;
import org.hibernate.Transaction;
import org.hibernate.criterion.Criterion;
import org.hibernate.criterion.Restrictions;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.data.MoodleUser;
import at.gv.egiz.moodle.eidconnector.webapp.util.WebAppUtil;
import at.iaik.commons.util.HibernateUtil;
import at.iaik.commons.util.MiscUtil;
import at.iaik.commons.util.ToStringUtil;

/**
 * @author AT-TUG (Thomas Knall)
 */
public final class MoodleUserAP {
   
   private static final Logger LOG = Logger.getLogger(MoodleUserAP.class);
   public static final String MOODLE_AUTH_TYPE_STORK = "stork";
   
   private MoodleUserAP() {
   }

   @SuppressWarnings("unchecked")
   public static List<MoodleUser> findUsersByCriterion(Criterion... criterions) {
      Transaction tx = null;
      List<MoodleUser> result = null;
      StopWatch sw = new StopWatch();
      sw.start();
      try {
         Session session = HibernateUtil.getCurrentSession(HBM_NAME_MOODLE);
         tx = session.beginTransaction();

         Criteria criteria = session.createCriteria(MoodleUser.class);

         if (MiscUtil.isNotEmpty(criterions)) {
            LOG.debug("Looking for user(s) (criterion(s) = " + ToStringUtil.toString(criterions) + ").");
            for (Criterion criterion : criterions) {
               criteria.add(criterion);
            }
         }
         LOG.debug("Effective criteria: " + criteria.toString());
         result = criteria.list();
         tx.commit();
         sw.stop();
         LOG.debug("Found " + result.size() + " user(s). Query time: " + sw.getTime() + "ms.");
      } catch (HibernateException hex) {
         LOG.error(hex);
         if (tx != null) {
            LOG.debug("Initiating rollback.");
            try {
               tx.rollback();
            } catch (HibernateException rollbackFailed) {
               LOG.error("Rollback failed (reason: " + rollbackFailed.getMessage() + ")");
            }
         }
         sw.stop();
         LOG.debug("Transaction aborted after " + sw.getTime() + "ms.");
         throw hex;
      }
      return result;
   }

   protected static MoodleUser findUserByCriterion(Criterion... criterions) {
      MiscUtil.assertNotEmpty(criterions, "Criterions");
      Transaction tx = null;
      MoodleUser result = null;
      StopWatch sw = new StopWatch();
      sw.start();
      try {
         Session session = HibernateUtil.getCurrentSession(HBM_NAME_MOODLE);
         tx = session.beginTransaction();

         Criteria criteria = session.createCriteria(MoodleUser.class);
         if (MiscUtil.isNotEmpty(criterions)) {
            LOG.debug("Looking for user (criterion(s) = " + ToStringUtil.toString(criterions) + ").");
            for (Criterion criterion : criterions) {
               criteria.add(criterion);
            }
         }
         LOG.debug("Effective criteria: " + criteria.toString());
         result = (MoodleUser) criteria.uniqueResult();
         tx.commit();
         sw.stop();
         LOG.debug((result != null ? "Found user." : "Unable to find user.") + " Query time: " + sw.getTime() + "ms.");
      } catch (HibernateException hex) {
         LOG.error(hex);
         if (tx != null) {
            LOG.debug("Initiating rollback.");
            try {
               tx.rollback();
            } catch (HibernateException rollbackFailed) {
               LOG.error("Rollback failed (reason: " + rollbackFailed.getMessage() + ")");
            }
         }
         sw.stop();
         LOG.debug("Transaction aborted after " + sw.getTime() + "ms.");
         throw hex;
      }
      return result;
   }

   
   public static MoodleUser findUserById(int id) {
      MiscUtil.assertNotNull(id, "id");
      return findUserByCriterion(Restrictions.eq("id", id));
   }
   
   public static MoodleUser findUserByUsername(String username) {
      MiscUtil.assertNotNull(username, "username");
      return findUserByCriterion(Restrictions.eq("username", username));
   }
   
   public static MoodleUser findSTORKUserByEIdentifier(String eIdentifier) throws NoSuchAlgorithmException, UnsupportedEncodingException {
      MiscUtil.assertNotNull(eIdentifier, "eIdentifier");
      String hexDigest = WebAppUtil.calculateDigestFromEIdentifier(eIdentifier);
      LOG.debug("Trying to find Moodle STORK user with username (derived from eIdentifier) \"" + hexDigest + "\".");
      return findUserByCriterion(Restrictions.eq("username", hexDigest), Restrictions.eq("auth", MOODLE_AUTH_TYPE_STORK));
   }
   
   public static MoodleUser findUserByUsernameAndPassword(String username, String password) {
      MiscUtil.assertNotNull(username, "username");
      MiscUtil.assertNotNull(username, "password");
      
      String[] configSalts = Configuration.getMoodleAccountSalts();
      if (configSalts == null) {
         configSalts = new String[] { };
      }
      ArrayList<String> saltsList = new ArrayList<String>(configSalts.length + 1);
      saltsList.add("");
      for (String salt : configSalts) {
         if (MiscUtil.isNotEmpty(salt)) {
            saltsList.add(salt);
         }
      }
      String oldCharset = Configuration.getMoodleOldCharset();
      String legacyPassword = null;
      if (MiscUtil.isNotEmpty(oldCharset)) {
         try {
            legacyPassword = new String(password.getBytes(oldCharset), "UTF-8");
         } catch (UnsupportedEncodingException e) {
            LOG.error("Error converting legacy password.", e);
         }
      }
      for (String salt : saltsList.toArray(new String[saltsList.size()])) {
         // calculate MD5-hash from password
         LOG.debug("Calculating message digest (" + Configuration.getMoodleProperty("moodle.hash.algorithm.password") + ") from password (salt = \"" + salt + "\").");
         byte[] digest = null;
         byte[] legacyDigest = null;
         try {
            MessageDigest md = MessageDigest.getInstance(Configuration.getMoodleProperty("moodle.hash.algorithm.password")); 
            digest = md.digest((password + salt).getBytes("UTF-8"));
            if (MiscUtil.isNotEmpty(legacyPassword)) {
               md.reset();
               legacyDigest = md.digest((legacyPassword + salt).getBytes("UTF-8"));
            }
         } catch (UnsupportedEncodingException e) {
            // could only happen if java installation is broken
            LOG.error(e);
         } catch (NoSuchAlgorithmException e) {
            // Has already been checked by configuration validator
            LOG.error(e);
         }
         String hexDigest = Util.toString(digest, "").toLowerCase();
         String legacyHexDigest = null;
         if (MiscUtil.isNotEmpty(legacyPassword)) {
            legacyHexDigest = Util.toString(legacyDigest, "").toLowerCase();
         }
         
         MoodleUser moodleUser = findUserByCriterion(Restrictions.eq("username", username), Restrictions.eq("hashOfPassword", hexDigest));
         if (moodleUser == null && MiscUtil.isNotEmpty(legacyHexDigest)) {
            LOG.debug("User not found. Trying legacy password version (" + oldCharset + ").");
            moodleUser = findUserByCriterion(Restrictions.eq("username", username), Restrictions.eq("hashOfPassword", legacyHexDigest));
         }
         if (moodleUser != null) {
            return moodleUser;
         }
      }
      return null;
      
   }
   
}
