/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.logging.impl;

import java.util.Date;

import org.apache.commons.lang.time.DateFormatUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.gv.egiz.moodle.eidconnector.logging.StatisticLogger;

/**
 * Implementation of a statistical logger based on slf4j that returns CSV based log entries.
 * Note that instead of commas, semicolons are used as delimiter.
 * @see http://tools.ietf.org/html/rfc4180
 * @author AT-TUG (Thomas Knall)
 */
public class CSVStatisticLogger implements StatisticLogger {
   
   public  final static String DELIMITER = ";";
   
   private final static String QUOTE = "\"";
   private final static String TIMESTAMP_PATTERN = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'";
   
   private Logger log = LoggerFactory.getLogger(CSVStatisticLogger.class);

   /**
    * Creates a log entry with log level DEBUG using the data provided.
    * Note that timestamp, assertionIssuerId and spID must not be {@code null}, while the remaining
    * parameters might be.<br/>
    * After parameter {@code age} an arbitrary number of various user defined parameters might be
    * passed. These additional fields are also logged if provided.<br/>
    * Values that are {@code null} are replaced by an empty String:<br/>
    * e.g. if {@code citizenCountryCode == null} and {@code age == null}<br/>
    *    {@code 2010-09-07T09:27:20.697Z;VIDP-AT;SaferChat-AT;eFKN39cIZoDjRRZOKhYlTD3dcXE=;;4;}
    * @param timestamp The current date. Should be formatted as UTC timestamp according to "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'".
    * @param assertionIssuerId The identifier of the underlying S-PEPS/V-IDP instance.
    * @param spID The identifier of the service provider being served.
    * @param sha1OfEID A base64 encoded String of the SHA1 hash of the citizen's eIdentifier.
    * @param citizenCountryCode The citizen's home country code (according to ISO 3166-1 alpha-2).
    * @param citizenQAALevel The underlying QAA level.
    * @param age The citizen's age.
    * @param others An arbitrary number of further fields to be logged.
    */
   public void log(Date timestamp, String assertionIssuerId, String spID, String sha1OfEID,
            String citizenCountryCode, Integer citizenQAALevel, Integer age, Object... others) {
      
      StringBuffer result = new StringBuffer();
      
      // timestamp
      result.append(prepareField(DateFormatUtils.formatUTC(timestamp, TIMESTAMP_PATTERN))).append(DELIMITER);
      
      // assertionIssuerId
      result.append(prepareField(assertionIssuerId)).append(DELIMITER);
      
      // spID
      result.append(prepareField(spID)).append(DELIMITER);
      
      // sha1OfEID
      result.append(prepareField(sha1OfEID)).append(DELIMITER);
      
      // citizenCountryCode
      result.append(prepareField(citizenCountryCode)).append(DELIMITER);
      
      // citizenQAALevel
      result.append(prepareField(citizenQAALevel)).append(DELIMITER);
      
      // age
      result.append(prepareField(age));
      
      // other
      if (others != null && others.length != 0) {
         for (Object other : others) {
            result.append(DELIMITER);
            if (other != null && other instanceof Date) {
               result.append(prepareField(DateFormatUtils.formatUTC(timestamp, TIMESTAMP_PATTERN)));               
            } else {
               result.append(prepareField(other));
            }
         }
      }
      
      log.debug(result.toString());
      
   }
   
   /**
    * Escapes fields if needed according to RFC 4180 otherwise the respective field is returned as is. 
    * @param field The field to be processed. {@code null} fields are returned as empty Strings.
    * @return The field, either escaped (with quotes) or not.
    * @see http://tools.ietf.org/html/rfc4180
    */
   private static String prepareField(Object field) {
      if (field == null) {
         return "";
      }
      
      // If double-quotes are used to enclose fields, then a double-quote
      // appearing inside a field must be escaped by preceding it with
      // another double quote.
      String value = String.valueOf(field).replaceAll(QUOTE, QUOTE + QUOTE);
      
      // Fields containing line breaks (CRLF), double quotes, and commas
      // should be enclosed in double-quotes.
      if (value.contains(DELIMITER) || value.contains("\n") || value.contains(QUOTE)) {
         return QUOTE + value + QUOTE;
      }
      return value;
   }
   
}
