/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.logging.impl;

import java.util.Date;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.data.StatisticLog;
import at.gv.egiz.moodle.eidconnector.logging.StatisticLogger;
import at.iaik.commons.util.HibernateUtil;

/**
 * Implementation of a statistical logger based on slf4j that returns CSV based log entries.
 * Note that instead of commas, semicolons are used as delimiter.
 * @see http://tools.ietf.org/html/rfc4180
 * @author AT-TUG (Thomas Knall)
 */
public class DatabaseStatisticLogger implements StatisticLogger {

   private Logger log = LoggerFactory.getLogger(DatabaseStatisticLogger.class);

   /**
    * Creates a log entry into the eid connector database. Note that the configuration key
    * {@code logger.database.enabled} in category {@code internal} needs to be set to {@code true}.
    * @param timestamp The current date.
    * @param assertionIssuerId The identifier of the underlying S-PEPS/V-IDP instance.
    * @param spID The identifier of the service provider being served.
    * @param sha1OfEID A base64 encoded String of the SHA1 hash of the citizen's eIdentifier.
    * @param citizenCountryCode The citizen's home country code (according to ISO 3166-1 alpha-2).
    * @param citizenQAALevel The underlying QAA level.
    * @param age The citizen's age.
    * @param others An arbitrary number of further fields to be logged.
    */
   public void log(Date timestamp, String assertionIssuerId, String spID, String sha1OfEID,
            String citizenCountryCode, Integer citizenQAALevel, Integer age, Object... others) {

      if (!Configuration.isDatabaseLoggingSupported()) {
         log.warn("Unable to log to database since database logging has not been enabled in configuration (configuration key \"logger.database.enabled\" in category \"internal\").");
         return;
      }

      StatisticLog sl = new StatisticLog(timestamp, assertionIssuerId, spID, sha1OfEID, citizenCountryCode, citizenQAALevel, age);

      // other
      if (others != null && others.length > 0) {
         if (others[0] != null) {
            sl.setCustomField1(others[0].toString());
         }
         if (others.length > 1 && others[1] != null) {
            sl.setCustomField2(others[1].toString());
         }
      }
      try {
         HibernateUtil.saveOrUpdate(Configuration.HBM_NAME_EID_CONNECTOR, sl);
      } catch (Exception e) {
         log.error("Error logging into database.", e);
      }

   }

}
