/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.validation;

import java.util.HashMap;
import java.util.Map;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.log4j.Logger;
import org.apache.xpath.XPathAPI;
import org.opensaml.common.SignableSAMLObject;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egovernment.moa.spss.MOAException;
import at.gv.egovernment.moa.spss.api.SPSSFactory;
import at.gv.egovernment.moa.spss.api.SignatureVerificationService;
import at.gv.egovernment.moa.spss.api.common.Content;
import at.gv.egovernment.moa.spss.api.xmlverify.VerifySignatureInfo;
import at.gv.egovernment.moa.spss.api.xmlverify.VerifySignatureLocation;
import at.gv.egovernment.moa.spss.api.xmlverify.VerifyXMLSignatureRequest;
import at.gv.egovernment.moa.spss.api.xmlverify.VerifyXMLSignatureResponse;
import eu.stork.mw.common.exceptions.SignatureVerificationException;
import eu.stork.mw.common.utils.StringUtils;
import eu.stork.mw.common.utils.XMLUtil;
import eu.stork.vidp.messages.exception.SAMLValidationException;

public class MOASPSSVerifier {

   private final static Logger log = Logger.getLogger(MOASPSSVerifier.class);

   private final String SIGNATURE_NAMESPACE_URI = "http://www.w3.org/2000/09/xmldsig#";
   private final String SIGNATURE_LOCAL_NAME = "Signature";

   public void verifySAMLObjectSignature(SignableSAMLObject samlObject)
         throws SAMLValidationException, SignatureVerificationException {

      log.debug("Verifying signature of " + samlObject.getElementQName());

      verifyXML(samlObject.getDOM(), true);

      log.info("Successful signature verification of " + samlObject.getElementQName());

   }

   public void verifyXML(Element signedXML, boolean certificateValidation)
         throws SignatureVerificationException {

      String path = Configuration.getSTORKResponseValidationSignatureVerificationConfigPath();
      System.setProperty("moa.spss.server.configuration", path);

      SPSSFactory spssFac = SPSSFactory.getInstance();
      SignatureVerificationService sigVerifyService = SignatureVerificationService.getInstance();

      Map<String, String> nSMap = getNamespaceMap(signedXML);

      String sigPrefix = null;
      NodeList nList = null;

      Element nsNode;
      try {
         nsNode = XMLUtil.createNewDocument().createElement("DummyNamespaceNode");
         nsNode.setAttribute("xmlns:ds", SIGNATURE_NAMESPACE_URI);

         // Assertion
         Node signatureNode = XPathAPI.selectSingleNode(signedXML, "//ds:Signature", nsNode);
         sigPrefix = signatureNode.getPrefix();
      } catch (Exception e1) {
         log.error("Cannot find signature element prefix.", e1);
         throw new SignatureVerificationException("Cannot find signature element prefix.", e1);
      }

      XPathFactory xpathFactory = XPathFactory.newInstance();
      XPath xpath = xpathFactory.newXPath();
      try {
         nList = (NodeList) xpath.evaluate(".", signedXML, XPathConstants.NODESET);
         // xpath.setNamespaceContext(nsContext)
      } catch (XPathExpressionException e) {
         throw new SignatureVerificationException("Cannot find node. ", e);
      }

      Content sigDocContent = spssFac.createContent(nList, null);

      String signatureXPathURI = StringUtils.isEmpty(sigPrefix) ? "//Signature" : "//" + sigPrefix
            + ":Signature";

      VerifySignatureLocation sigLocation = spssFac.createVerifySignatureLocation(
            signatureXPathURI, nSMap);
      VerifySignatureInfo sigInfo = spssFac.createVerifySignatureInfo(sigDocContent, sigLocation);

      String trustProfileId = Configuration
            .getSTORKResponseValidationSignatureVerificationProfile();

      VerifyXMLSignatureRequest verifyRequest = spssFac.createVerifyXMLSignatureRequest(null, // Validation
                                                                                              // time
            sigInfo, null, // no supplements
            null, // no verification of signature manifest
            false, // do not return hash input
            trustProfileId); // TrustProfileID

      VerifyXMLSignatureResponse verifyResponse = null;
      try {
         // calling MOA-SP
         verifyResponse = sigVerifyService.verifyXMLSignature(verifyRequest);
      } catch (MOAException e) {
         log.error(e.getMessage());
         e.printStackTrace();
         throw new SignatureVerificationException("Error invoking MOA-SP", e);
      }

      if (verifyResponse.getSignatureCheck().getCode() != 0)
         throw new SignatureVerificationException("Could not validate digital signature");

      log.debug("Signature succesfully verified");

      if (certificateValidation) {
         if (verifyResponse.getCertificateCheck().getCode() != 0)
            throw new SignatureVerificationException("Could not validate signing certificate");
         log.debug("Certificate succesfully verified");
      }

      log.info("Successful signature verification.");
   }

   private Map<String, String> getNamespaceMap(Element element) {

      Map<String, String> nsMap = new HashMap<String, String>();

      addNamespacesOfNode(element, nsMap);

      return nsMap;
   }

   private void addNamespacesOfNode(Node node, Map<String, String> map) {

      String prefix = node.getPrefix();
      String nsUri = node.getNamespaceURI();

      if (!map.containsKey(prefix)) {

         map.put(prefix, nsUri);
      }

      NodeList children = node.getChildNodes();

      for (int i = 0; i < children.getLength(); i++) {

         Node current = children.item(i);

         addNamespacesOfNode(current, map);
      }
   }

}
