/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.validation;

import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.Audience;
import org.opensaml.saml2.core.AudienceRestriction;
import org.opensaml.saml2.core.Conditions;
import org.opensaml.saml2.core.SubjectConfirmation;
import org.opensaml.saml2.core.SubjectConfirmationData;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;

/**
 * @author bzwattendorfer
 *
 */
public class PEPSConnectorAssertionVerifier implements AssertionVerifier {
   
   private final Logger log = Logger.getLogger(this.getClass().getName());
   
   private static final int CLOCK_SCEW_MINUTES = 5;

   public void verify(Assertion assertion, String reqIPAddress, String authnRequestID, String recipient) throws SecurityException {

      // SAML assertion need not to be signed, skipping signature validation
      verifyIssuer(assertion.getIssuer().getValue());
      verifySubjectConfirmation(assertion, reqIPAddress, authnRequestID, recipient);
      
      String audience = Configuration.getSTORKResponseValidationExpectedAudience();
      
      // FIXME: remove
//      verifyConditions(assertion, audience);
      
   }
   
   private void verifyIssuer(String issuer) throws SecurityException {
      if (!Configuration.isSTORKResponseValidationTrustedIssuer(issuer)) {
         log.error("Issuer \"" + issuer + "\" not trusted.");
         throw new SecurityException("Issuer not trusted.");
      }
      log.trace("Issuer \"" + issuer + "\" is trusted.");
   }
   
   private void verifySubjectConfirmation(Assertion assertion, String reqAddress, String requestID, String recipient) throws SecurityException {      
      for (SubjectConfirmation sc : assertion.getSubject().getSubjectConfirmations()) {
         verifySubjectConfirmationData(sc.getSubjectConfirmationData(), reqAddress, requestID, recipient);
      }
            
   }
   
   private void verifySubjectConfirmationData(SubjectConfirmationData scData, String reqAddress, String requestID, String recipient) throws SecurityException {            
      //NotBefore not allowed in SSO profile
      
      // FIXME: remove
//      verifyNotOnOrAfter(scData.getNotOnOrAfter());
      // FIXME: remove
//      verifyRecipient(scData, recipient);
      
      
//      verifyInResponseTo(scData, requestID);
             
   }
   
   private void verifyNotBefore(DateTime notBefore) throws SecurityException {
      if (notBefore.minusMinutes(CLOCK_SCEW_MINUTES).isAfterNow()) {
         String msg = "Subject/Assertion not yet valid, Timestamp: " + notBefore;
         log.error(msg);
         throw new SecurityException(msg);
      }
      log.trace("Subject/Assertion already valid, notBefore: " + notBefore);
      
   }
   
   private void verifyNotOnOrAfter(DateTime notOnOrAfter) throws SecurityException {
      if (notOnOrAfter.plusMinutes(CLOCK_SCEW_MINUTES).isBeforeNow()) {
         String msg = "Subject/Assertion no longer valid.";
         log.error(msg);
         throw new SecurityException(msg);
      }
      
      log.trace("Subject/Assertion still valid, notOnOrAfter: " + notOnOrAfter);
   }

   private void verifyClientAddress(SubjectConfirmationData scData, String reqAddress) throws SecurityException {
      if (!reqAddress.equals(scData.getAddress())) {
         String msg = "Response coming from wrong Client-Address";
         log.error("Response coming from wrong Client-Address " + reqAddress + " but expected " + scData.getAddress());
         throw new SecurityException(msg);
      }
      
   }
   
   private void verifyInResponseTo(SubjectConfirmationData scData, String requestID) throws SecurityException {
      if (!scData.getInResponseTo().equals(requestID)) {
         String msg = "Assertion issued for wrong request";
         log.error(msg);
         throw new SecurityException(msg);
      }
   }
   
   private void verifyRecipient(SubjectConfirmationData scData, String reqRecipient) throws SecurityException {
      if (!scData.getRecipient().equals(reqRecipient)) {
         String msg = "Assertion intended for another recipient";
         log.error("Assertion intended for recipient \"" + scData.getRecipient() + "\" but expected \"" + reqRecipient + "\".");
         throw new SecurityException(msg);
      }
      
   }
   
   private void verifyAudience(AudienceRestriction audienceRestriction, String reqAudience) throws SecurityException {
      for (Audience audience : audienceRestriction.getAudiences()) {
         if (audience.getAudienceURI().equals(reqAudience))
            return;
      }
      String msg = "Assertion sent to wrong audience";
      log.error("Assertion intended for wrong audience, expected " + reqAudience);
      throw new SecurityException(msg);
   }
   
   private void verifyConditions(Assertion assertion, String reqAudience) throws SecurityException {     
      Conditions conditions = assertion.getConditions();
      
      verifyNotBefore(conditions.getNotBefore());
      verifyNotOnOrAfter(conditions.getNotOnOrAfter());
      
      verifyAudience(conditions.getAudienceRestrictions().get(0), reqAudience);
      
   }
   
}
