/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.validation;

import org.apache.log4j.Logger;

import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import eu.stork.mw.common.exceptions.SecurityException;
import eu.stork.mw.common.exceptions.SignatureVerificationException;
import eu.stork.mw.messages.saml.STORKResponse;
import eu.stork.vidp.messages.exception.SAMLValidationException;
import eu.stork.vidp.messages.util.SAMLUtil;

/**
 * @author bzwattendorfer
 *
 */
public class PEPSConnectorResponseVerifier implements ResponseVerifier {
   
   private final Logger log = Logger.getLogger(this.getClass().getName());
      
   /* (non-Javadoc)
    * @see eu.stork.mw.peps.connector.validation.ResponseVerifier#verify(org.opensaml.saml2.core.Response)
    */
   public void verify(STORKResponse response) throws SecurityException {
      
      verifySignature(response);
      log.trace("Signature of SAML response valid.");
      
      verifyStandardValidation(response);
      
      log.trace("SAML response format valid.");
      
      verifyIssuer(response.getIssuer().getValue());
      
      log.trace("Issuer of SAML response is known: " + response.getIssuer().getValue());
   }
   
   
   private void verifySignature(STORKResponse response) throws SecurityException {
      //validate Signature
      try {
         if (response.isSigned()) {          
            MOASPSSVerifier verifier = new MOASPSSVerifier();          
            verifier.verifySAMLObjectSignature(response);
         } else {
            String msg = "SAML Response is not signed.";
            throw new SecurityException(msg);
         }
         
      } catch (SAMLValidationException e) {
         String msg = "Error validating SAMLResponse from PEPS";
         log.error(msg, e);
         throw new SecurityException(msg, e);
      } catch (SignatureVerificationException e) {
         String msg = "Error validating Signature of SAMLResponse";
         log.error(msg, e);
         throw new SecurityException(msg, e);
      }           
      
   }
   
   private void verifyStandardValidation(STORKResponse response) throws SecurityException {
      try {
         SAMLUtil.verifySAMLObjectStandardValidation(response, "saml2-core-schema-and-stork-validator");
      } catch (SAMLValidationException e) {
         String msg ="SAML Response received not valid.";
         throw new SecurityException(msg, e);
      }
      
   }
   
   private void verifyIssuer(String issuer) throws SecurityException {
      if (!Configuration.isSTORKResponseValidationTrustedIssuer(issuer)) {
         throw new SecurityException("Issuer not trusted.");
      }
   }
   
}

