/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.webapp.struts.action;

import iaik.utils.Util;

import java.io.IOException;
import java.io.Writer;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.util.Date;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.lang.BooleanUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.commons.lang.time.DateFormatUtils;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.gv.egiz.moodle.eidconnector.EIDConnectorException;
import at.gv.egiz.moodle.eidconnector.auth.AuthenticationProvider;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributes;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributesStore;
import at.gv.egiz.moodle.eidconnector.auth.impl.AuthenticationImplProvider;
import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.data.EIDMoodleUserMapping;
import at.gv.egiz.moodle.eidconnector.data.IdMapping;
import at.gv.egiz.moodle.eidconnector.data.MoodleUser;
import at.gv.egiz.moodle.eidconnector.data.ap.EIDMoodleUserMappingAP;
import at.gv.egiz.moodle.eidconnector.data.ap.IdMappingAP;
import at.gv.egiz.moodle.eidconnector.data.ap.MoodleUserAP;
import at.gv.egiz.moodle.eidconnector.logging.StatisticLogger;
import at.gv.egiz.moodle.eidconnector.logging.StatisticLoggerProvider;
import at.gv.egiz.moodle.eidconnector.webapp.struts.Constants;
import at.gv.egiz.moodle.eidconnector.webapp.util.WebAppUtil;
import at.iaik.commons.util.CryptUtil;
import at.iaik.commons.util.DateTimeUtil;
import at.iaik.commons.util.HibernateUtil;
import at.iaik.commons.util.MiscUtil;
import at.iaik.commons.util.ToStringUtil;
import at.iaik.commons.xmlschema.saml.AgeOverData;

/**
 * @author AT-TUG (Thomas Knall)
 */
public class MoodleEIDConnectorAction extends CommonAction {

   private final Logger log = LoggerFactory.getLogger(MoodleEIDConnectorAction.class);

   public static final String ORIGINATING_COUNTRY_PARAM = "country";
   public static final String TICKET_PARAM = "ticket";
   public static final int    INVALID_TICKET_RETURN_CODE             = HttpServletResponse.SC_FORBIDDEN; 
   public static final int    NO_TICKET_RETURN_CODE                  = HttpServletResponse.SC_BAD_REQUEST; 
   public static final int    INSUFFICIENT_ATTRIBUTES_RETURN_CODE    = HttpServletResponse.SC_INTERNAL_SERVER_ERROR; 
   public static final int    UNABLE_TO_FIND_USER_RETURN_CODE        = HttpServletResponse.SC_BAD_REQUEST; 
   public static final int    UNPARSEABLE_USER_ID_RETURN_CODE        = HttpServletResponse.SC_INTERNAL_SERVER_ERROR; 
   public static final int    RETURNING_CREDENTIALS_RETURN_CODE      = HttpServletResponse.SC_OK; 
   public static final String RETURNING_CREDENTIALS_CONTENT_TYPE     = "text/xml"; 
   public static final String RETURNING_CREDENTIALS_CONTENT_CHAR_ENC = "UTF-8"; 
   public static final String RETURNING_CREDENTIALS_CONTENT          = "<ok/>"; 

   protected static final String SUBJECT_ATTRIBUTE_KEY_MOODLE_USER_ID = "http://www.moodle.org/userId";
   
   public ActionForward startAuthentication(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws NoSuchAlgorithmException, IOException {
      
      if (Configuration.isMaintenanceMode()) {
         log.info("StartAuthentication attempted but rejected since eID connector is in maintenance mode.");
         return this.prepareMessagePage(mapping, request, "errors.maintenancemode", Constants.FWD_LEAVE_CONNECTOR);
      }

      // get originating country code (defined in ISO 3166-1 alpha-2)
      String country = StringUtils.trimToEmpty(request.getParameter(ORIGINATING_COUNTRY_PARAM)).toUpperCase();

      // check if parameter for originating country was provided
      if (MiscUtil.isEmpty(country)) {
         return this.prepareMessagePage(mapping, request, "errors.nocountrycode", null, ORIGINATING_COUNTRY_PARAM);
      }
      
      // check if country code provided is valid
      boolean stork = Configuration.CONFIG_STORK_AUTHENTICATOR_ID.equalsIgnoreCase(country);
      if (!stork && !MiscUtil.isValidCountryCode(country)) {
         return this.prepareMessagePage(mapping, request, "errors.invalidcountrycode", null, country);
      }
      Locale currentBrowserLocale = super.getLocale(request);
      String localizedCountryName = stork ? Configuration.CONFIG_STORK_AUTHENTICATOR_ID : (new Locale("", country)).getDisplayCountry(currentBrowserLocale);
      
      // check if a country specific interface has been configured
      String startAuthURL = at.iaik.commons.util.WebAppUtil.substituteHttpServletRequestData(Configuration.getAuthCountryProperty(country, "startauthentication.url"), request);
      
      if (MiscUtil.isEmpty(startAuthURL) || !Configuration.isAuthCountryEnabled(country)) {
         return this.prepareMessagePage(mapping, request, "errors.notsupportedcountry", null, country, localizedCountryName);
      }
      
      // adding service recognition header
      WebAppUtil.addServiceHeader(response);
      log.debug("Redirecting to " + startAuthURL);
      response.sendRedirect(response.encodeRedirectURL(startAuthURL));
      return null;
   }

   /**
    * Action that has to be invoked by the underlying authentication framework. This action performs
    * delegation to the set of registered eID authenticators.
    */
   public ActionForward authenticate(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws NoSuchAlgorithmException, IOException {

      if (Configuration.isMaintenanceMode()) {
         log.info("StartAuthentication attempted but rejected since eID connector is in maintenance mode.");
         return this.prepareMessagePage(mapping, request, "errors.maintenancemode", Constants.FWD_LEAVE_CONNECTOR);
      }

      String[] registeredCountries = AuthenticationImplProvider.getInstance().getAuthenticatedCountries();
      Locale currentBrowserLocale = super.getLocale(request);
      for (String countryCode : registeredCountries) {
         boolean stork = Configuration.CONFIG_STORK_AUTHENTICATOR_ID.equalsIgnoreCase(countryCode);
         String localizedCountryName = stork ? Configuration.CONFIG_STORK_AUTHENTICATOR_ID : (new Locale("", countryCode)).getDisplayCountry(currentBrowserLocale);
         try {
            AuthenticationProvider ap = AuthenticationImplProvider.getInstance().getAuthenticationProvider(countryCode);
            if (ap.isResponsible(request)) {
               SubjectAttributes sa = null;
               try {
                  sa = ap.authenticate(request);
               } catch (Exception e) {
                  String stackTrace = ExceptionUtils.getFullStackTrace(e);
                  log.error(stackTrace);
//                  log.error(e);
                  if (Configuration.getInternalBooleanProperty("show.authenticator.errors")) {
                     return this.prepareMessagePage(mapping, request, "errors.authprovider", Constants.FWD_LEAVE_CONNECTOR, localizedCountryName, e.getMessage());
                  }
                  log.debug("Trying next authenticator in queue.");
                  continue;
               }
               if (sa != null) {
                  log.debug("Subject attributes received from Authentication Provider [" + countryCode + "].");
                  log.trace("Authentication data: " + sa);
                  
                  // check if eid was provided
                  if (MiscUtil.isEmpty(sa.getEIdentifier())) {
                     return this.prepareMessagePage(mapping, request, "errors.noeidprovided", Constants.FWD_LEAVE_CONNECTOR, localizedCountryName);
                  }
                  
                  // check if date of birth has been set (try to calculate if not set)
                  if (sa.getAge() == null) {
                     log.debug("Age has not been set.");
                     if (sa.getDateOfBirth() != null) {
                        log.debug("Date of birth given. Calculating age.");
                        if (sa.getValidAt() == null) {
                           log.debug("No specific date set for validity of attributes. Assuming current date.");
                           sa.setAgeValue(DateTimeUtil.calcAge(sa.getDateOfBirthValue(), new Date()));
                        } else {
                           log.debug("Date assuring validity of attributes at " + DateFormatUtils.ISO_DATETIME_TIME_ZONE_FORMAT.format(sa.getValidAt()) + " provided.");
                           sa.setAgeValue(DateTimeUtil.calcAge(sa.getDateOfBirthValue(), sa.getValidAt()));
                        }
                     } else {
                        log.info("Date of birth has not been provided. Unable to calculate age.");
                     }
                  }
                  
                  if (!sa.isAlreadyLogged()) {
                     StatisticLoggerProvider.getInstance().log(
                              new Date(),
                              ap.getClass().getName(),
                              Configuration.getMoodleProperty("moodle.ticket.consumer.service"),
                              MiscUtil.isNotEmpty(sa.getEIdentifier()) ? CryptUtil.calcBase64Hash(sa.getEIdentifier().getBytes(), "SHA") : null,
                              sa.get(SubjectAttributes.NATIONALITY),
                              sa.getQAALevelValue(),
                              sa.getAgeValue()
                     );
                     sa.setAlreadyLogged(true);
                  }
                  
                  MoodleUser moodleUser = null;
                  
                  // check if a moodle account mapping already exists
                  EIDMoodleUserMapping eIDMoodleMapping = EIDMoodleUserMappingAP.findMappingByEIdentifier(sa.getEIdentifier());
                  
                  if (Configuration.isEIdTransformationSupported()) {
                     // eid-transformation-helper: if no eID mapping found try mapping via givenName/familyName/dateOfBirth
                     if (eIDMoodleMapping == null && MiscUtil.areAllNotEmpty(sa.getGivenName(), sa.getSurName()) && sa.getDateOfBirthValue() != null) {
                        log.debug("eid-transformation-helper: eIDMoodleMapping not found. Trying to find id mapping for a potentially old identifier.");
                        IdMapping idMapping = IdMappingAP.findIdMappingByPersonData(sa.getGivenName(), sa.getSurName(), sa.getDateOfBirthValue());
                        if (idMapping != null && !sa.getEIdentifier().equals(idMapping.getNameIdentifier())) {
                           log.debug(
                                    "eid-transformation-helper: Id mapping of old identifier \"{}\" ({}) -> new identifier \"{}\" ({}) found.",
                                    new Object[] {
                                             idMapping.getNameIdentifier(), idMapping.getNameQualifier(),
                                             sa.getEIdentifier(), sa.get(SubjectAttributes.EQUALIFIER)
                                    }
                           );
                           log.debug("eid-transformation-helper: Trying to find eId mapping for old identifier.");
                           eIDMoodleMapping = EIDMoodleUserMappingAP.findMappingByEIdentifier(idMapping.getNameIdentifier());
                           if (eIDMoodleMapping != null) {
                              log.debug("eid-transformation-helper: eID mapping for old identifier found. Updating eID mapping and removing id mapping.");
                              eIDMoodleMapping.setEIdentifier(sa.getEIdentifier());
                              eIDMoodleMapping.setEQualifier(sa.get(SubjectAttributes.EQUALIFIER));
                              HibernateUtil.saveOrUpdateDelete(
                                       Configuration.HBM_NAME_EID_CONNECTOR,
                                       new EIDMoodleUserMapping[] { eIDMoodleMapping },
                                       new IdMapping[] { idMapping }
                              );
                           }
                        } else {
                           log.debug("eid-transformation-helper: No old id mapping for this person found.");
                        }
                     }
                  }
                  
                  if (eIDMoodleMapping == null) {
                     log.debug("No Moodle user account currently mapped to eID. Checking if new account has just been created.");
                     
                     // check if a new account has been created for this user (therefore no mapping is currently available)
                     // for new accounts a username derived from eIdentifier is used
                     moodleUser = MoodleUserAP.findSTORKUserByEIdentifier(sa.getEIdentifier());
                     
                     if (Configuration.isEIdTransformationSupported()) {
                        if (moodleUser == null && MiscUtil.areAllNotEmpty(sa.getGivenName(), sa.getSurName()) && sa.getDateOfBirthValue() != null) {
                           log.debug("eid-transformation-helper: Moodle user account for current eid not found. Trying to find id mapping for old identifier.");
                           IdMapping idMapping = IdMappingAP.findIdMappingByPersonData(sa.getGivenName(), sa.getSurName(), sa.getDateOfBirthValue());
                           if (idMapping != null && !sa.getEIdentifier().equals(idMapping.getNameIdentifier())) {
                              log.debug(
                                       "eid-transformation-helper: Id mapping of old identifier \"{}\" ({}) -> new identifier \"{}\" ({}) found.",
                                       new Object[] {
                                                idMapping.getNameIdentifier(), idMapping.getNameQualifier(),
                                                sa.getEIdentifier(), sa.get(SubjectAttributes.EQUALIFIER)
                                       }
                              );
                              log.debug("eid-transformation-helper: Trying to find newly created Moodle user account for old identifier.");
                              moodleUser = MoodleUserAP.findSTORKUserByEIdentifier(idMapping.getNameIdentifier());
                              if (moodleUser != null) {
                                 log.debug("eid-transformation-helper: Newly created Moodle user account for old identifier found. Removing id mapping.");
                                 HibernateUtil.delete(Configuration.HBM_NAME_EID_CONNECTOR, idMapping);
                              }
                           } else {
                              log.debug("eid-transformation-helper: No old id mapping for this person found.");
                           }
                        }
                     }
                     
                     if (moodleUser == null) {
                        log.debug("No new account has been created.");
                        
                        // check if we need a registration dialogue (for selection of registration type)
                        if (Configuration.isMoodleExistingAccountsEnabled()) {
                           request.getSession().setAttribute(Constants.SESS_SUBJECT_ATTRIBUTES, sa);
                           log.debug("Redirecting to registration.");
                           return mapping.findForward(Constants.FWD_NEW_REGISTRATION);
                        } else {
                           log.debug("Support for existing accounts is disabled. Skipping registration dialogue.");
                           String ticket = SubjectAttributesStore.getInstance().storeSubjectAttributes(request.getSession().getId(), sa);
                           String moodleURL = response.encodeRedirectURL(Configuration.getMoodleProperty("moodle.ticket.consumer.service"));
                           URL redirectURL = at.iaik.commons.util.WebAppUtil.appendPostParameter(new URL(moodleURL), new NameValuePair("ticket", ticket));
                           // adding service recognition header
                           WebAppUtil.addServiceHeader(response);
                           log.debug("Ticket will be sent to " + moodleURL);
                           log.debug("Redirecting to " + redirectURL);
                           response.sendRedirect(redirectURL.toString());
                           return null;
                        }
                        
                     } else {
                        log.debug("No account mapped but newly created account discovered. Storing eId mapping.");
                        // creating mapping
                        eIDMoodleMapping = new EIDMoodleUserMapping();
                        eIDMoodleMapping.setEIdentifier(sa.getEIdentifier());
                        eIDMoodleMapping.setEQualifier(sa.get(SubjectAttributes.EQUALIFIER));
                        log.debug("Saving moodle user account mapping: " + sa.getEIdentifier() + " -> " + moodleUser.getUsername() + " (id=" + moodleUser.getId() + ").");
                        eIDMoodleMapping.setMoodleUserId(moodleUser.getId());
                        HibernateUtil.saveOrUpdate(Configuration.HBM_NAME_EID_CONNECTOR, eIDMoodleMapping);
                        log.debug("Successfully saved.");
                     }
                  }
                  if (eIDMoodleMapping != null) {

                     // eID-Moodle mapping available.
                     log.debug("eID-Moodle mapping available.");
                     
                     if (Configuration.isEIdTransformationSupported()) {
                        // eid-transformation-helper: if person data is available store it together with eid in order to be able to convert id on qualifier change
                        if (MiscUtil.areAllNotEmpty(sa.getGivenName(), sa.getSurName()) && sa.getDateOfBirthValue() != null) {
                           // check if id mapping has already been stored
                           log.debug("eid-transformation-helper: Looking if id mapping already exists.");
                           IdMapping idMapping = IdMappingAP.findIdMappingByPersonData(sa.getGivenName(), sa.getSurName(), sa.getDateOfBirthValue());
                           if (idMapping == null) {
                              log.debug("eid-transformation-helper: Id mapping not found. Storing new mapping for later eid transformation.");
                              idMapping = new IdMapping(sa.getGivenName(), sa.getSurName(), sa.getDateOfBirthValue());
                              idMapping.setNameIdentifier(sa.getEIdentifier());
                              idMapping.setNameQualifier(sa.get(SubjectAttributes.EQUALIFIER));
                              HibernateUtil.saveOrUpdate(Configuration.HBM_NAME_EID_CONNECTOR, idMapping);
                           } else {
                             log.debug("eid-transformation-helper: Id mapping has already been stored before.");
                           }
                        }
                     }
                     
                     // update eQualifier if not already set.
                     if (MiscUtil.isEmpty(eIDMoodleMapping.getEQualifier()) && (MiscUtil.isNotEmpty(sa.get(SubjectAttributes.EQUALIFIER)))) {
                        log.debug("Updating eQualifier.");
                        eIDMoodleMapping.setEQualifier(sa.get(SubjectAttributes.EQUALIFIER));
                        HibernateUtil.saveOrUpdate(Configuration.HBM_NAME_EID_CONNECTOR, eIDMoodleMapping);
                        log.debug("Successfully saved.");
                     }

                     // check if Moodle user is really available
                     moodleUser = MoodleUserAP.findUserById(eIDMoodleMapping.getMoodleUserId());
                     if (moodleUser == null) {
                        log.error("There was a mapping (" + eIDMoodleMapping + "\" for the given eID, but no valid Moodle account was associated.");
                        return this.prepareMessagePage(mapping, request, "errors.nomoodleaccountassociated", Constants.FWD_LEAVE_CONNECTOR, localizedCountryName);
                     }
                     
                  }
                  
                  // check if Moodle account has been deleted
                  if (BooleanUtils.isTrue(moodleUser.getDeleted())) {
                     log.info("Moodle account mapping found, but account has been deleted.");
                     request.getSession().setAttribute(Constants.SESS_SUBJECT_ATTRIBUTES, sa);
                     if (eIDMoodleMapping != null) {
                        log.debug("Removing account mapping.");
                        HibernateUtil.delete(Configuration.HBM_NAME_EID_CONNECTOR, eIDMoodleMapping);
                     }
                     // check if we need a registration dialogue (for selection of registration type)
                     if (Configuration.isMoodleExistingAccountsEnabled()) {
                        log.debug("Redirecting to registration.");
                        return mapping.findForward(Constants.FWD_NEW_REGISTRATION);
                     } else {
                        log.debug("Support for existing accounts is disabled. Skipping registration dialogue.");
                        String ticket = SubjectAttributesStore.getInstance().storeSubjectAttributes(request.getSession().getId(), sa);
                        String moodleURL = response.encodeRedirectURL(Configuration.getMoodleProperty("moodle.ticket.consumer.service"));
                        URL redirectURL = at.iaik.commons.util.WebAppUtil.appendPostParameter(new URL(moodleURL), new NameValuePair("ticket", ticket));
                        // adding service recognition header
                        WebAppUtil.addServiceHeader(response);
                        log.debug("Ticket will be sent to " + moodleURL);
                        log.debug("Redirecting to " + redirectURL);
                        response.sendRedirect(redirectURL.toString());
                        return null;
                     }
                  }
                  
                  // check if Moodle account's auth type is within the list of allowed types
                  String[] allowedAuthTypes = Configuration.getMoodleArrayProperty("moodle.allow.auth.types");
                  String formattedAllowedAuthTypes = ToStringUtil.toString(allowedAuthTypes, ", ", "\"");
                  if (!MiscUtil.inIgnoreCase(moodleUser.getAuth(), allowedAuthTypes)) {
                     log.warn("Invalid auth type of Moodle user account " + moodleUser + ". Must be one of [" + formattedAllowedAuthTypes + "].");
                     return this.prepareMessagePage(mapping, request, "errors.invalidmoodleauthtype", Constants.FWD_LEAVE_CONNECTOR, formattedAllowedAuthTypes);
                  }

                  // put moodle user id to subject attributes for temporal storage
                  sa.set(SUBJECT_ATTRIBUTE_KEY_MOODLE_USER_ID, String.valueOf(moodleUser.getId()));
                  
                  String ticket = SubjectAttributesStore.getInstance().storeSubjectAttributes(request.getSession().getId(), sa);
                  String moodleURL = response.encodeRedirectURL(Configuration.getMoodleProperty("moodle.ticket.consumer.service"));
                  URL redirectURL = at.iaik.commons.util.WebAppUtil.appendPostParameter(new URL(moodleURL), new NameValuePair("ticket", ticket));
                  // adding service recognition header
                  WebAppUtil.addServiceHeader(response);
                  log.debug("Ticket will be sent to " + moodleURL);
                  log.debug("Redirecting to " + redirectURL);
                  response.sendRedirect(redirectURL.toString());
                  return null;
               } else {
                  log.error("No attributes provided by authenticator for " + countryCode + ".");
                  if (Configuration.getInternalBooleanProperty("show.authenticator.errors")) {
                     return this.prepareMessagePage(mapping, request, "errors.noattributesprovided", null, localizedCountryName);
                  }
                  // better give the next authenticator a try
                  log.debug("Trying next authenticator in queue.");
                  continue;
               }
            }
         } catch (EIDConnectorException e) {
            String stackTrace = ExceptionUtils.getFullStackTrace(e);
            log.error(stackTrace);
//            log.error(e);
            return this.prepareMessagePage(mapping, request, "errors.authprovider", Constants.FWD_LEAVE_CONNECTOR, localizedCountryName, e.getMessage());
         }
      }
      
      return this.prepareMessagePage(mapping, request, "errors.noauthmethodfound", Constants.FWD_LEAVE_CONNECTOR);
   }
   
   public ActionForward getAuthenticationData(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws IOException, NoSuchAlgorithmException {
      String ticket = StringUtils.trimToNull(request.getParameter(TICKET_PARAM));
      if (ticket == null) {
         log.warn("Missing parameter \"" + TICKET_PARAM + "\".");
         response.sendError(NO_TICKET_RETURN_CODE);
         return null;
      }
      
      SubjectAttributes sa = SubjectAttributesStore.getInstance().getSubjectAttributes(ticket);
      if (sa == null) {
         log.warn("No SubjectAttributes found for ticket \"" + ticket + "\".");
         response.sendError(INVALID_TICKET_RETURN_CODE);
         return null;
      }
      log.debug("Ticket \"" + ticket + "\" accepted.");
      
      // search for moodle user
      String moodleUserIdStr = sa.get(SUBJECT_ATTRIBUTE_KEY_MOODLE_USER_ID);
      if (MiscUtil.areAllEmpty(moodleUserIdStr, sa.getEIdentifier())) {
         log.debug("SubjectAttributes do neither contain Moodle user id nor eIdentifier.");
         response.sendError(INSUFFICIENT_ATTRIBUTES_RETURN_CODE);
         return null;
      }
      
      String moodleUsername = null;
      if (MiscUtil.isNotEmpty(moodleUserIdStr)) {
         int moodleUserId;
         try {
            moodleUserId = Integer.parseInt(moodleUserIdStr);
         } catch (NumberFormatException e) {
            log.error("Unable to parse given Moodle user id \"" + moodleUserIdStr + "\".", e);
            response.sendError(UNPARSEABLE_USER_ID_RETURN_CODE);
            return null;
         }
         MoodleUser moodleUser = MoodleUserAP.findUserById(moodleUserId);
         if (moodleUser == null) {
            log.error("Unable to find Moodle user account.");
            response.sendError(UNABLE_TO_FIND_USER_RETURN_CODE);
            return null;
         } else {
            
            // user has been found
            if (BooleanUtils.isTrue(moodleUser.getDeleted())) {
               log.error("Moodle user account found, but has been deleted.");
               response.sendError(UNABLE_TO_FIND_USER_RETURN_CODE);
               return null;
            }
            log.debug("Moodle user account found: " + moodleUser);
            moodleUsername = moodleUser.getUsername();
         }
      } else {
         // account is going to be created by moodle, so we do not yet have user data from database
         log.debug("Account should be created by Moodle. Deriving username from given eIdentifier.");
         moodleUsername = WebAppUtil.calculateDigestFromEIdentifier(sa.getEIdentifier());
      }

      log.debug("Adding mapped header(s).");
      // process mapped header attributes directly from stored SubjectAttributes
      Map<String, String> httpHeaderMappings = Configuration.getEIDHTTPHeaderMappings();
      Iterator<String> it = sa.getKeySet().iterator();
      while (it.hasNext()) {
         String uri = it.next();
         String value = sa.get(uri);
         String httpHeader = httpHeaderMappings.get(uri);
         if (value != null) {
            if (MiscUtil.isNotEmpty(httpHeader)) {
               String base64Value = Util.toBase64String(value.getBytes("UTF-8"));
               log.debug("Adding http header \"" + httpHeader + "\" with value \"" + base64Value + "\" (= \"" + value + "\") taken from \"" + uri + "\".");
               response.addHeader(httpHeader, base64Value);
            } else {
               log.trace("No http header mapping defined for attribute \"" + uri + "\".");
            }
         }
      }
      // is age over
      for (AgeOverData ageOver : sa.getIsAgeOver()) {
         String uri = SubjectAttributes.IS_AGE_OVER;
         String httpHeader = httpHeaderMappings.get(uri);
         String value = "";
         if (ageOver.getAge() != null) {
            value = String.valueOf(ageOver.getAge()) + ":";
         }
         value += String.valueOf(ageOver.isFulfilled());
         if (MiscUtil.isNotEmpty(httpHeader)) {
            String base64Value = Util.toBase64String(value.getBytes("UTF-8"));
            log.debug("Adding http header \"" + httpHeader + "\" with value \"" + base64Value + "\" (= \"" + value + "\") taken from \"" + uri + "\".");
            response.addHeader(httpHeader, base64Value);
         } else {
            log.debug("No http header mapping defined for attribute \"" + uri + "\".");
         }
      }
      
      String moodleUsernameHeader = Configuration.getMoodleProperty("moodle.httpheader.username");
      String moodleUserNameBase64 = Util.toBase64String(moodleUsername.getBytes("UTF-8"));
      log.debug("Adding Moodle user name http header \"" + moodleUsernameHeader + "\" = \"" + moodleUserNameBase64 + "\" (= \"" + moodleUsername + "\").");
      response.addHeader(moodleUsernameHeader, moodleUserNameBase64);
      
      // adding service recognition header
      WebAppUtil.addServiceHeader(response);
      
      response.setStatus(RETURNING_CREDENTIALS_RETURN_CODE);
      response.setContentType(RETURNING_CREDENTIALS_CONTENT_TYPE);
      response.setCharacterEncoding(RETURNING_CREDENTIALS_CONTENT_CHAR_ENC);
      String content = RETURNING_CREDENTIALS_CONTENT;
      response.setContentLength(content.getBytes(RETURNING_CREDENTIALS_CONTENT_CHAR_ENC).length);
      Writer output = response.getWriter();
      output.write(content);
      output.flush();
      output.close();
      
      return null;
   }
   
   public ActionForward leaveConnector(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws IOException {
      log.debug("Leaving connector.");
      HttpSession httpSession = request.getSession();
      WebAppUtil.cleanUpSession(httpSession);

      String redirectURL = Configuration.getMoodleProperty("moodle.start.page");
      log.debug("Redirecting to " + redirectURL);
      response.sendRedirect(redirectURL.toString());
      return null;
   }

}
