/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */
package at.gv.egiz.moodle.eidconnector.webapp.struts.action;

import java.io.IOException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.httpclient.NameValuePair;
import org.apache.commons.lang.BooleanUtils;
import org.apache.log4j.Logger;
import org.apache.struts.action.ActionErrors;
import org.apache.struts.action.ActionForm;
import org.apache.struts.action.ActionForward;
import org.apache.struts.action.ActionMapping;
import org.apache.struts.action.ActionMessage;
import org.apache.struts.action.ActionMessages;

import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributes;
import at.gv.egiz.moodle.eidconnector.auth.data.SubjectAttributesStore;
import at.gv.egiz.moodle.eidconnector.core.config.Configuration;
import at.gv.egiz.moodle.eidconnector.data.EIDMoodleUserMapping;
import at.gv.egiz.moodle.eidconnector.data.MoodleUser;
import at.gv.egiz.moodle.eidconnector.data.ap.EIDMoodleUserMappingAP;
import at.gv.egiz.moodle.eidconnector.data.ap.MoodleUserAP;
import at.gv.egiz.moodle.eidconnector.webapp.struts.Constants;
import at.gv.egiz.moodle.eidconnector.webapp.struts.RegistrationType;
import at.gv.egiz.moodle.eidconnector.webapp.struts.form.RegistrationForm;
import at.gv.egiz.moodle.eidconnector.webapp.util.WebAppUtil;
import at.iaik.commons.util.HibernateUtil;
import at.iaik.commons.util.MiscUtil;
import at.iaik.commons.util.ToStringUtil;

/**
 * @author AT-TUG (Thomas Knall)
 */
public class RegistrationAction extends CommonAction {

   private final Logger log = Logger.getLogger(RegistrationAction.class);

   public ActionForward init(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws NoSuchAlgorithmException, IOException {
      log.debug("Preparing registration form.");
      HttpSession session = request.getSession();
      if (session.getAttribute(Constants.SESS_SUBJECT_ATTRIBUTES) == null) {
         WebAppUtil.cleanUpSession(session);
         return this.prepareMessagePage(mapping, request, "errors.nosubjectattributesfound", Constants.FWD_LEAVE_CONNECTOR);
      }
      
      RegistrationForm rf = (RegistrationForm) form;
      rf.setRegistrationType(RegistrationType.NEW_ACCOUNT.name());
      return mapping.getInputForward();
   }

   public ActionForward updateForm(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) {
      HttpSession session = request.getSession();
      if (session.getAttribute(Constants.SESS_SUBJECT_ATTRIBUTES) == null) {
         WebAppUtil.cleanUpSession(session);
         return this.prepareMessagePage(mapping, request, "errors.nosubjectattributesfound", Constants.FWD_LEAVE_CONNECTOR);
      }
      return mapping.getInputForward();
   }

   
   public ActionForward submit(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) throws NoSuchAlgorithmException, IOException {
      log.trace("Validating registration form.");

      HttpSession session = request.getSession();
      SubjectAttributes sa = (SubjectAttributes) session.getAttribute(Constants.SESS_SUBJECT_ATTRIBUTES);
      if (sa == null) {
         return this.prepareMessagePage(mapping, request, "errors.nosubjectattributesfound", Constants.FWD_LEAVE_CONNECTOR);
      }

      // invoke validation
      ActionMessages messages = new ActionMessages();
      RegistrationForm registrationForm = (RegistrationForm) form;
      registrationForm.trim();
      ActionErrors actionErrors = registrationForm.validate(mapping, request); 
      messages.add(actionErrors);
      
      boolean createNewAccount = RegistrationType.NEW_ACCOUNT.name().equals(registrationForm.getRegistrationType());
      if (createNewAccount && !Configuration.isMoodleNewAccountsEnabled()) {
         return this.prepareMessagePage(mapping, request, "errors.notallowed", Constants.FWD_LEAVE_CONNECTOR);
      }
      
      // do user account validation here because we will need the user object later on
      // check if profile already exists
      // since check is time consuming, do check only if no other errors are present
      MoodleUser moodleUser = null;
      if (actionErrors.isEmpty() && !createNewAccount) {
         moodleUser = MoodleUserAP.findUserByUsernameAndPassword(registrationForm.getMoodleUserName(), registrationForm.getMoodlePassword());
         if ((moodleUser == null) || (BooleanUtils.isTrue(moodleUser.getDeleted()))) {
            log.debug("No active Moodle user account with specific username/password was found.");
            messages.add("moodleUserName", new ActionMessage("errors.invalidmoodlecredentials"));
            messages.add("moodlePassword", new ActionMessage("", false));
         } else {
            // check if Moodle account's auth type is within the list of allowed types
            String[] allowedAuthTypes = Configuration.getMoodleArrayProperty("moodle.allow.auth.types");
            String formattedAllowedAuthTypes = ToStringUtil.toString(allowedAuthTypes, ", ", "\"");
            if (!MiscUtil.inIgnoreCase(moodleUser.getAuth(), allowedAuthTypes)) {
               log.warn("Invalid auth type of Moodle user account " + moodleUser + ". Must be one of [" + formattedAllowedAuthTypes + "].");
               messages.add(ActionMessages.GLOBAL_MESSAGE, new ActionMessage("errors.invalidmoodleauthtype", formattedAllowedAuthTypes));
            }
         }
         
      }
      
      // abort on error(s)
      if (!messages.isEmpty()) {
         log.debug("Validation failed.");
         this.saveErrors(request, messages);
         return mapping.getInputForward();
      }

      // validation ok, proceed
      log.debug("Validation ok.");
      
      if (!createNewAccount) {
         // check if we have to update an existing mapping or if we store a new mapping
         EIDMoodleUserMapping eIDMoodleMapping = EIDMoodleUserMappingAP.findMappingByEIdentifier(sa.getEIdentifier());
         if (eIDMoodleMapping == null) {
            log.debug("No existing mapping found. Creating a new mapping.");
            eIDMoodleMapping = new EIDMoodleUserMapping();
            eIDMoodleMapping.setEIdentifier(sa.getEIdentifier());
            eIDMoodleMapping.setEQualifier(sa.get(SubjectAttributes.EQUALIFIER));
         } else {
            log.debug("Mapping already exists. Updating mapping.");
         }
         
         log.debug("Saving moodle user account mapping: " + sa.getEIdentifier() + " -> " + moodleUser.getUsername() + " (id=" + moodleUser.getId() + ").");
         eIDMoodleMapping.setMoodleUserId(moodleUser.getId());
         HibernateUtil.saveOrUpdate(Configuration.HBM_NAME_EID_CONNECTOR, eIDMoodleMapping);
         log.debug("Successfully saved.");
         
         if (!MoodleUserAP.MOODLE_AUTH_TYPE_STORK.equals(moodleUser.getAuth())) {
            moodleUser.setAuth(MoodleUserAP.MOODLE_AUTH_TYPE_STORK);
            log.debug("Updating moodle user data (setting auth type to stork): " + moodleUser);
            HibernateUtil.update(Configuration.HBM_NAME_MOODLE, moodleUser);
            log.debug("Successfully updated.");
         }

         // put moodle user id to subject attributes
         sa.set(MoodleEIDConnectorAction.SUBJECT_ATTRIBUTE_KEY_MOODLE_USER_ID, String.valueOf(moodleUser.getId()));
      }

      String ticket = SubjectAttributesStore.getInstance().storeSubjectAttributes(request.getSession().getId(), sa);
      String moodleURL = response.encodeRedirectURL(Configuration.getMoodleProperty("moodle.ticket.consumer.service"));
      URL redirectURL = at.iaik.commons.util.WebAppUtil.appendPostParameter(new URL(moodleURL), new NameValuePair("ticket", ticket));
      // adding service recognition header
      WebAppUtil.addServiceHeader(response);
      log.debug("Ticket will be sent to " + moodleURL);
      log.debug("Redirecting to " + redirectURL);
      response.sendRedirect(redirectURL.toString());
      return null;
   }

   @Override
   protected ActionForward cancelled(ActionMapping mapping, ActionForm form, HttpServletRequest request, HttpServletResponse response) {
      return mapping.findForward(Constants.FWD_LEAVE_CONNECTOR);
   }

}
