<?php
/*
 * Copyright 2010 Graz University of Technology
 *
 * Licensed under the EUPL, Version 1.1 or - as soon they will be approved by
 * the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 * http://www.osor.eu/eupl/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 *
 * This product combines work with different licenses. See the "NOTICE" text
 * file for details on the various modules and licenses.
 * The "NOTICE" text file is part of the distribution. Any derivative works
 * that you distribute must include a readable copy of the "NOTICE" text file.
 */

 /**
 * @author Thomas Knall
 * @license http://www.osor.eu/eupl/ EUPL Version 1.1
 * @package moodle multiauth
 *
 * Authentication Plugin: STORK Authentication
 *
 * Authentication using STORK eID Modules based on the Shibboleth authentication plugin by
 * Martin Dougiamas and Lukas Haemmerle
 *
 * 2009-11-10  first release
 */

if (!defined('MOODLE_INTERNAL')) {
    die('Direct access to this script is forbidden.');    ///  It must be included from a Moodle page
}

require_once($CFG->libdir.'/authlib.php');

/**
 * STORK authentication plugin.
 */
class auth_plugin_stork extends auth_plugin_base {

    /**
     * Constructor.
     */
    function auth_plugin_stork() {
        $this->authtype = 'stork';
        $this->config = get_config('auth/stork');
    }

    /**
     * Returns true if the username and password work and false if they are
     * wrong or don't exist.
     *
     * @param string $username The username (with system magic quotes)
     * @param string $password The password (with system magic quotes)
     * @return bool Authentication success or failure.
     */
    function user_login($username, $password) {

        $stork = $_SESSION['STORK_SESSION'];
        // check if everything we need is in session and valid
        return (isset($stork) && isset($stork['username']) && ($stork['username'] == $username));
       
    }

    /**
     * Returns array containg attribute mappings between Moodle and STORK skipping empty values.
     *
     * @return array
     */
    function get_attributes() {
        $configarray = (array) $this->config;

        $moodleattributes = array();
        foreach ($this->userfields as $field) {
            if (isset($configarray["field_map_$field"]) && !empty($configarray["field_map_$field"])) {
                $moodleattributes[$field] = $configarray["field_map_$field"];
            }
        }
        $moodleattributes['username'] = $configarray["user_attribute"];

        return $moodleattributes;
    }
    
    /**
     * Read user information from external database and returns it as array().
     *
     * @param string $username username (with system magic quotes)
     *
     * @return mixed array with no magic quotes or false on error
     */
    function get_userinfo($username) {
        $stork = $_SESSION['STORK_SESSION'];
        if (isset($stork)) {
        	return $stork;
        } else {
        	return array();
        }
    }
    
    /**
     * Returns true if this authentication plugin is 'internal'.
     *
     * @return bool
     */
    function is_internal() {
        return false;
    }

    /**
     * Returns true if this authentication plugin can change the user's
     * password.
     *
     * @return bool
     */
    function can_change_password() {
        return false;
    }

     /**
     * Hook for login page
     *
     */
    function loginpage_hook() {
        global $SESSION, $CFG;

        // Prevent username from being shown on login page after logout
        $CFG->nolastloggedin = true;

        return;
    }
    
    /**
     * Prints a form for configuring this authentication plugin.
     *
     * This function is called from admin/auth.php, and outputs a full page with
     * a form for configuring this plugin.
     *
     * @param array $page An object containing all the data for this page.
     */
    function config_form($config, $err, $user_fields) {
        include "config.html";
    }

    /**
     * Processes and stores configuration data for this authentication plugin.
     *
     *
     * @param object $config Configuration object
     */
    function process_config($config) {
        global $CFG;

        // set to defaults if undefined
        if (!isset($config->auth_instructions) or empty($config->user_attribute)) {
            $config->auth_instructions = get_string('stork_instructions', 'auth', $CFG->wwwroot.'/auth/stork/index.php');
        }
        if (!isset ($config->user_attribute)) {
            $config->user_attribute = 'X-STORK-Moodle-username';
        }
        if (!isset ($config->age_attribute)) {
            $config->age_attribute = 'X-STORK-age';
        }
        if (!isset ($config->qaalevel_attribute)) {
            $config->qaalevel_attribute = 'X-STORK-qaaLevel';
        }
        if (!isset ($config->connector_url)) {
            $config->connector_url = '';
        }
        
        if (!isset($config->login_name)) {
            $config->login_name = 'STORK Login';
        }

        if (!isset ($config->min_qaalevel)) {
            $config->min_qaalevel = '0';
        }

        // save settings
        set_config('user_attribute',     $config->user_attribute,     'auth/stork');
        set_config('age_attribute',      $config->age_attribute,      'auth/stork');
        set_config('qaalevel_attribute', $config->qaalevel_attribute, 'auth/stork');
        set_config('min_qaalevel',       $config->min_qaalevel,       'auth/stork');
        set_config('login_name',         $config->login_name,         'auth/stork');
        set_config('connector_url',      $config->connector_url,      'auth/stork');
        

        return true;
    }
        
}
    
?>
